"""
Unrelated module from the step definition test app.
"""

# If this module is imported by Aloe, it will cause an error.
raise Exception("This module should not be loaded when searching for steps.")



# -*- coding: utf-8 -*-
"""
Test parsing steps.
"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()


I_LIKE_VEGETABLES = '''
Given I hold a special love for green vegetables
'''

I_HAVE_TASTY_BEVERAGES = '''
Given I have the following tasty beverages in my freezer:
   | Name   | Type     | Price |
   | Skol   | Beer     |  3.80 |
   | Nestea | Ice-tea  |  2.10 |
'''

I_DIE_HAPPY = '''
Given I shall die with love in my heart
'''

BACKGROUND_WITH_TAGGED_SCENARIO = '''
    Background:
        background line 1

    @wip
    Scenario:
        Scenario line 1
'''

MULTI_LINE = '''
Given I have a string like so:
  """
  This is line one
  and this is line two
  and this is line three
    and this is line four,

    with spaces at the beginning
  """
'''.strip()


from nose.tools import assert_equal

from aloe.parser import Feature, Step


def parse_steps(step):
    """Parse a step, prefixing it with a feature and a scenario header."""
    feature = """
    Feature: parse a step
    Scenario: parse a single step
    """

    feature += step

    return Feature.from_string(feature).scenarios[0].steps


def first_line_of(step):
    """
    Return the first line of a step
    """
    return step.strip().splitlines()[0]


def test_step_has_repr():
    """
    Step implements __repr__ nicely
    """
    step, = parse_steps(I_HAVE_TASTY_BEVERAGES)
    assert_equal(
        repr(step),
        '<Step: "' + first_line_of(I_HAVE_TASTY_BEVERAGES) + '">'
    )


def test_can_get_sentence_from_string():
    """
    It should extract the sentence string from the whole step
    """

    step, = parse_steps(I_HAVE_TASTY_BEVERAGES)

    assert isinstance(step, Step)

    assert_equal(
        step.sentence,
        first_line_of(I_HAVE_TASTY_BEVERAGES)
    )


def test_can_parse_keys_from_table():
    """
    It should take the keys from the step, if it has a table
    """

    step, = parse_steps(I_HAVE_TASTY_BEVERAGES)
    assert_equal(step.keys, ('Name', 'Type', 'Price'))


def test_can_parse_tables():
    """
    It should have a list of data from a given step, if it has a table
    """

    step, = parse_steps(I_HAVE_TASTY_BEVERAGES)

    assert_equal(step.hashes, (
        {
            'Name': 'Skol',
            'Type': 'Beer',
            'Price': '3.80'
        },
        {
            'Name': 'Nestea',
            'Type': 'Ice-tea',
            'Price': '2.10'
        },
    ))


def test_can_parse_a_unary_array_from_single_step():
    """
    It should extract a single ordinary step correctly into an array of steps
    """

    steps = parse_steps(I_HAVE_TASTY_BEVERAGES)
    assert_equal(len(steps), 1)
    assert isinstance(steps[0], Step)
    assert_equal(steps[0].sentence,
                 first_line_of(I_HAVE_TASTY_BEVERAGES))


def test_can_parse_a_unary_array_from_complicated_step():
    """
    It should extract a single tabular step correctly into an array of steps
    """

    steps = parse_steps(I_LIKE_VEGETABLES)
    assert_equal(len(steps), 1)
    assert isinstance(steps[0], Step)
    assert_equal(steps[0].sentence, first_line_of(I_LIKE_VEGETABLES))


def test_can_parse_regular_step_followed_by_tabular_step():
    """
    It should correctly extract two steps (one regular, one tabular) into an
    array.
    """
    steps = parse_steps(I_LIKE_VEGETABLES + I_HAVE_TASTY_BEVERAGES)
    assert_equal(len(steps), 2)
    assert isinstance(steps[0], Step)
    assert isinstance(steps[1], Step)
    assert_equal(steps[0].sentence, first_line_of(I_LIKE_VEGETABLES))
    assert_equal(steps[1].sentence, first_line_of(I_HAVE_TASTY_BEVERAGES))


def test_can_parse_tabular_step_followed_by_regular_step():
    """"
    It should correctly extract two steps (one tabular, one regular) into
    an array.
    """

    steps = parse_steps(I_HAVE_TASTY_BEVERAGES + I_LIKE_VEGETABLES)
    assert_equal(len(steps), 2)
    assert isinstance(steps[0], Step)
    assert isinstance(steps[1], Step)
    assert_equal(steps[0].sentence, first_line_of(I_HAVE_TASTY_BEVERAGES))
    assert_equal(steps[1].sentence, first_line_of(I_LIKE_VEGETABLES))


def test_can_parse_two_ordinary_steps():
    """
    It should correctly extract two ordinary steps into an array.
    """

    steps = parse_steps(I_DIE_HAPPY + I_LIKE_VEGETABLES)
    assert_equal(len(steps), 2)
    assert isinstance(steps[0], Step)
    assert isinstance(steps[1], Step)
    assert_equal(steps[0].sentence, first_line_of(I_DIE_HAPPY))
    assert_equal(steps[1].sentence, first_line_of(I_LIKE_VEGETABLES))


def test_multiline_is_part_of_previous_step():
    """
    It should correctly parse a multi-line string as part of the preceding step
    """

    steps = parse_steps(MULTI_LINE)
    assert_equal(len(steps), 1)
    assert isinstance(steps[0], Step)
    assert_equal(steps[0].sentence, 'Given I have a string like so:')


def test_table_escaping():
    """
    Table columns can be correctly escaped
    """

    steps = parse_steps(r"""
    Given I have items in my table:
        | Column 1                 |
        | This is a column         |
        | This is \| also a column |
        | This is \\ a backslash   |
    """)

    assert_equal(len(steps), 1)

    step, = steps

    assert_equal(step.table, (
        (r'Column 1',),
        (r'This is a column',),
        (r'This is | also a column',),
        (r'This is \ a backslash',),
    ))


def test_multiline_is_parsed():
    """Test parsing a multiline string in a step."""
    step, = parse_steps(MULTI_LINE)
    assert_equal(step.sentence, 'Given I have a string like so:')
    assert_equal(step.multiline, u"""This is line one
and this is line two
and this is line three
  and this is line four,

  with spaces at the beginning""")


def test_step_with_hash():
    """Test parsing a step with a hash"""
    step, = parse_steps('''
    Given I have product #1234 in my cart
    ''')

    assert_equal(step.sentence, 'Given I have product #1234 in my cart')


def test_comments():
    """Test parsing Gherkin comments."""
    steps = parse_steps('''
    # A comment
    Given I have a comment
                        # Another comment
    And I have another comment
    ''')

    assert_equal(
        [step.sentence for step in steps],
        [
            'Given I have a comment',
            'And I have another comment',
        ]
    )


def test_multiline_with_hash():
    """Test parsing a multiline with a hash"""
    step, = parse_steps('''
    Given I have the following products in my cart:
    """
    #1234
    #2345
    """
    ''')

    assert_equal(step.multiline, u"#1234\n#2345")



# -*- coding: utf-8 -*-
"""
Basic scenario tests.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()

import sys
import os
from inspect import getsourcefile

from aloe import world

from aloe.testing import (
    FeatureTest,
    in_directory,
)
from aloe.utils import PY3, TestWrapperIO


@in_directory('tests/simple_app')
class SimpleScenarioTest(FeatureTest):
    """
    Test that basic feature running works as expected.
    """

    def test_success(self):
        """
        Test running a simple feature.
        """

        self.assert_feature_success('features/calculator.feature')

    def test_success_zh(self):
        """
        Test running a simple feature in Chinese.
        """

        self.assert_feature_success('features/calculator_zh.feature')

    def test_failure(self):
        """
        Test that a failing feature fails tests.
        """

        self.assert_feature_fail('features/wrong_expectations.feature')

    def test_error_message(self):
        """
        Check that the appropriate error messages are printed on failure.
        """

        stream = TestWrapperIO()

        failing_feature = 'features/wrong_expectations.feature'

        self.assert_feature_fail(failing_feature, '-n', '1', stream=stream)

        output = stream.getvalue()

        error_header = "FAIL: Fail at adding " + \
            "(features.wrong_expectations: Wrong expectations)"

        self.assertIn(error_header, output)

        feature_stack_frame = """
  File "{feature}", line 11, in Fail at adding
    Then the result should be 40 on the screen
        """.strip().format(feature=os.path.abspath(failing_feature))

        self.assertIn(feature_stack_frame, output)

        step_file = getsourcefile(sys.modules['features.steps'])

        step_stack_frame = """
  File "{step_file}", line 64, in assert_result
    assert world.result == float(result)
AssertionError
        """.strip().format(step_file=step_file)

        self.assertIn(step_stack_frame, output)

    def test_failure_zh(self):
        """
        Test that a failing feature in Chinese fails tests.
        """

        stream = TestWrapperIO()

        failing_feature = 'features/wrong_expectations_zh.feature'

        self.assert_feature_fail(failing_feature, stream=stream)

        # Check that the appropriate error messages were printed

        output = stream.getvalue()

        error_header = "FAIL: 添加两个数值 " + \
            "(features.wrong_expectations_zh: 不对的预期)"

        self.assertIn(error_header, output)

        if PY3:
            feature_stack_frame = """
  File "{feature}", line 12, in 添加两个数值
    那么结果应该是40
            """.strip().format(feature=os.path.abspath(failing_feature))

            self.assertIn(feature_stack_frame, output)
        else:
            # Cannot use non-ASCII method names in Python 2
            feature_stack_frame = """
  File "{feature}", line 12, in
            """.strip().format(feature=os.path.abspath(failing_feature))

            self.assertIn(feature_stack_frame, output)

            feature_code_line = "那么结果应该是40"
            self.assertIn(feature_code_line, output)

        step_file = getsourcefile(sys.modules['features.steps'])

        step_stack_frame = """
  File "{step_file}", line 64, in assert_result
    assert world.result == float(result)
AssertionError
        """.strip().format(step_file=step_file)

        self.assertIn(step_stack_frame, output)

    def test_background(self):
        """
        Test running a scenario with a background.
        """

        self.assert_feature_success('features/background.feature')

    def test_outlines(self):
        """
        Test a scenario with outlines.
        """

        self.assert_feature_success('features/outlines.feature')

    def test_error_message_outline(self):
        """
        Check that the appropriate error messages are printed on failure in a
        scenario outline.
        """

        stream = TestWrapperIO()

        failing_feature = 'features/wrong_expectations.feature'

        self.assert_feature_fail(failing_feature, '-n', '2', stream=stream)

        output = stream.getvalue()

        error_header = "FAIL: Fail repeatedly " + \
            "(features.wrong_expectations: Wrong expectations)"

        self.assertIn(error_header, output)

        # TODO: Why isn't the line indented with 6 spaces like in the file?
        feature_stack_frame = """
  File "{feature}", line 22, in Fail repeatedly
    | 50     |
        """.strip().format(feature=os.path.abspath(failing_feature))

        self.assertIn(feature_stack_frame, output)

        example_stack_frame = """
  File "{feature}", line 18, in Fail repeatedly
    Then the result should be <result> on the screen
        """.strip().format(feature=os.path.abspath(failing_feature))

        self.assertIn(example_stack_frame, output)

        step_file = getsourcefile(sys.modules['features.steps'])

        step_stack_frame = """
  File "{step_file}", line 64, in assert_result
    assert world.result == float(result)
AssertionError
        """.strip().format(step_file=step_file)

        self.assertIn(step_stack_frame, output)

    def test_python_test_skipped(self):
        """
        Test that the Python test does not get picked up.
        """

        self.assert_feature_success('tests')

    def test_scenario_indices(self):
        """
        Test specifying the scenario indices to run.
        """

        feature = 'features/scenario_index.feature'

        # Run a single scenario
        self.assert_feature_success(feature, '-n', '1')
        self.assertEqual(world.all_results, [10])

        # Run multiple scenarios
        # TODO: Handling multi-valued options in Nose is tricky
        self.assert_feature_success(feature, '-n', '1,2')
        self.assertEqual(world.all_results, [10, 20])

        # Run a scenario outline
        self.assert_feature_success(feature, '-n', '3')
        self.assertEqual(world.all_results, [30, 40])

    def test_tags(self):
        """
        Test specifying the tags to run.
        """

        feature_dir = 'features/with_tags'
        feature_one = 'features/with_tags/one.feature'

        # Run scenarios by tag
        self.assert_feature_success(feature_one, '-a', 'hana')
        self.assertEqual(world.all_results, [1, 11])

        # Run scenarios with a tag that's not there
        self.assert_feature_success(feature_one, '-a', 'set')
        self.assertEqual(world.all_results, [])

        # Run features and scenarios with the tag
        self.assert_feature_success(feature_dir, '-a', 'dul')
        self.assertEqual(world.all_results, [2, 13, 20, 200])

        # Specify a tag to exclude
        self.assert_feature_success(feature_one, '-a', '!hana')
        self.assertEqual(world.all_results, [2, 4])

        # Specify more than one tag to run
        self.assert_feature_success(feature_one, '-a', 'hana', '-a', 'dul')
        self.assertEqual(world.all_results, [1, 2, 11])

        # Specify more than one tag to exclude
        self.assert_feature_success(feature_dir, '-a', '!hana,!dul')
        self.assertEqual(world.all_results, [4])



"""
Useful tools for writing Aloe steps.

See also :class:`aloe.world`.
"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

# pylint:disable=redefined-builtin, unused-wildcard-import, wildcard-import
from builtins import *
# pylint:enable=redefined-builtin, unused-wildcard-import, wildcard-import
from future import standard_library
standard_library.install_aliases()

import re
import inspect
from datetime import datetime
from functools import wraps


def guess_types(data):  # pylint:disable=too-complex
    """
    Converts a record or list of records from strings contained in
    outlines, table or hashes into a version with the types guessed.

    :param data: a :attr:`Scenario.outlines`, :attr:`Step.table`,
        :attr:`Step.hashes` or any other :class:`list`,
        list of lists or list of dicts.

    Will guess the following (in priority order):

     * :class:`bool` (``true``/``false``)
     * :class:`None` (``null``)
     * :class:`int`
     * :class:`date` in ISO format (``yyyy-mm-dd``)
     * :class:`str`

    The function operates recursively, so you should be able to pass nearly
    anything to it. At the very least basic types plus
    :class:`dict` and iterables.
    """

    # convert bytes to strings
    if isinstance(data, bytes):
        data = data.decode()

    if isinstance(data, str):
        if data == "true":
            data = True
        elif data == "false":
            data = False
        elif data == "null":
            data = None
        elif data.isdigit() and not re.match("^0[0-9]+", data):
            data = int(data)
        elif re.match(r'^\d{4}-\d{2}-\d{2}$', data):
            data = datetime.strptime(data, "%Y-%m-%d").date()
        else:
            # it's a string
            pass

        return data

    # if it's a dict, recurse as a dict
    if isinstance(data, dict):
        return {
            guess_types(key): guess_types(data)
            for key, data in data.items()
        }

    # try to recurse as an iterable
    try:
        return [guess_types(val) for val in data]
    except TypeError:
        pass

    #  give up
    return data


def hook_not_reentrant(func):
    """
    Decorate a hook as unable to be reentered while it is already in the
    stack.

    Any further attempts to enter the hook before exiting will be replaced
    by a no-op.

    This is generally useful for step hooks where a step might call
    :meth:`Step.behave_as` and trigger a second level of step hooks i.e.
    when displaying information about the running test.
    """

    # closure-scoped variable to track whether this hook has been
    # entered
    entered = [False]

    @wraps(func)
    def inner(*args, **kwargs):
        """
        Wrap func to check if we've already entered this function and if
        so replace it with a no-op.
        """

        def generator():
            """
            Hide the generator in a separate function
            because Python 2 can't support "returning from generators"
            """
            if entered[0]:
                yield
            else:
                try:
                    entered[0] = True
                    for val in func(*args, **kwargs):
                        yield val
                finally:
                    entered[0] = False

        if inspect.isgeneratorfunction(func):
            return generator()
        else:
            if not entered[0]:
                try:
                    entered[0] = True
                    return func(*args, **kwargs)
                finally:
                    entered[0] = False

    return inner



# -*- coding: utf-8 -*-
"""
Steps for testing the basic Gherkin test functionality.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()

from aloe import after, before, step, world

# pylint:disable=unused-argument


@before.all
def init_all_results():
    """Initialise the result storage."""
    world.all_results = []


@before.each_example
def clear(*args):
    """Clean the results for each example."""
    world.numbers = []
    world.result = 0


@step(r'I have entered (\d+) into the calculator')
@step(r'我输入(\d+)')
def enter_number(self, number):
    """Store the entered number."""
    world.numbers.append(float(number))


@step(r'I press add')
@step(r'我按添加')
def press_add(self):
    """Sum up the numbers."""
    world.result = sum(world.numbers)


@step(r'I press \[\+\]')
def press_plus(self):
    """
    Alias of 'I press add'.

    Tests behave_as().
    """
    self.given('I press add')


@step(r'I have a table')
def have_table(self):
    """Nothing."""


@step(r'The result should be (\d+) on the screen')
@step(r'结果应该是(\d+)')
def assert_result(self, result):
    """Assert the result is correct."""
    assert world.result == float(result)


@after.each_example
def record_all_results(scenario, outline, steps):
    """
    Record all results for the example
    """

    world.all_results.append(world.result)



"""
Test level 3 outputter
"""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
# pylint:disable=redefined-builtin, unused-wildcard-import, wildcard-import
from builtins import *
# pylint:enable=redefined-builtin, unused-wildcard-import, wildcard-import
from future import standard_library
standard_library.install_aliases()

from aloe.testing import (
    FeatureTest,
    in_directory,
)
from aloe.result import Terminal
from aloe.utils import TestWrapperIO
from mock import patch


class MockTermElement(object):
    """
    Wraps an element being output by the MockTerminal (i.e. term.bold) and
    replaces it with a string that can be checked in tests.

    Like with the attributes in Terminal it can be used as a callable or
    a string.
    """
    def __init__(self, attr):
        self.attr = attr

    def __call__(self, str_):
        # strips the string to make the output easier to read
        return 't.{attr}({str})'.format(attr=self.attr,
                                        str=str_.strip())

    def __str__(self):
        return '<t.{attr}>'.format(attr=self.attr)

    def __mul__(self, other):
        return str(self) * other


class MockTerminal(Terminal):
    """Mock terminal to output printable elements instead of ANSI sequences"""

    def __getattr__(self, attr):
        return MockTermElement(attr)

    def color(self, color):
        return getattr(self, 'color%s' % color)


@in_directory('tests/simple_app')
class OutputterTest(FeatureTest):
    """
    Test level 3 outputter
    """

    maxDiff = None

    def test_uncolored_output(self):
        """Test streamed output"""

        stream = TestWrapperIO()

        with patch('aloe.result.AloeTestResult.printSummary'):
            self.run_features('features/highlighting.feature',
                              verbosity=3, stream=stream)

            self.assertEqual(stream.getvalue(), """
Feature: Highlighting

  As a programmer
  I want to see my scenarios with pretty highlighting
  So my output is easy to read

  Scenario: behave_as works                            features/highlighting.feature:13
    Given I have a table
    Given I have entered 10 into the calculator
    And I press [+]

  Scenario Outline: Scenario outlines                  features/highlighting.feature:17
      | number |
      | 30     |

    Given I have a table
    Given I have entered 30 into the calculator
    And I press add

  Scenario Outline: Scenario outlines                  features/highlighting.feature:17
      | number |
      | 40     |

    Given I have a table
    Given I have entered 40 into the calculator
    And I press add

  @tables
  Scenario: Scenario with table                        features/highlighting.feature:27
    Given I have a table
    Given I have a table:
      | value |
      | 1     |
      | 1     |
      | 2     |

  Scenario: Scenario with a multiline                  features/highlighting.feature:34
    Given I have a table
    Given I have a table:
      \"\"\"
      Not actually a table :-P
      \"\"\"

""".lstrip())  # noqa

    def test_color_output(self):
        """Test streamed output with color"""

        stream = TestWrapperIO()

        with \
                patch('aloe.result.Terminal', new=MockTerminal), \
                patch('aloe.result.AloeTestResult.printSummary'):
            self.run_features('features/highlighting.feature',
                              verbosity=3, stream=stream,
                              force_color=True)

            self.assertEqual(stream.getvalue(), """
t.bold_white(Feature: Highlighting)

t.white(As a programmer
  I want to see my scenarios with pretty highlighting
  So my output is easy to read)

t.bold_white(Scenario: behave_as works)t.color8(features/highlighting.feature:13)
t.bold_green(Given I have a table)
t.bold_green(Given I have entered 10 into the calculator)
t.bold_green(And I press [+])

t.bold_white(Scenario Outline: Scenario outlines)t.color8(features/highlighting.feature:17)
      | t.white(number) |
      | t.white(30) |

t.bold_green(Given I have a table)
t.bold_green(Given I have entered 30 into the calculator)
t.bold_green(And I press add)

t.bold_white(Scenario Outline: Scenario outlines)t.color8(features/highlighting.feature:17)
      | t.white(number) |
      | t.white(40) |

t.bold_green(Given I have a table)
t.bold_green(Given I have entered 40 into the calculator)
t.bold_green(And I press add)

t.cyan(@tables)
t.bold_white(Scenario: Scenario with table)t.color8(features/highlighting.feature:27)
t.bold_green(Given I have a table)
t.bold_green(Given I have a table:)
      | t.bold_green(value) |
      | t.bold_green(1) |
      | t.bold_green(1) |
      | t.bold_green(2) |

t.bold_white(Scenario: Scenario with a multiline)t.color8(features/highlighting.feature:34)
t.bold_green(Given I have a table)
t.bold_green(Given I have a table:)
      \"\"\"
      t.bold_green(Not actually a table :-P)
      \"\"\"

""".lstrip())  # noqa

    def test_tty_output(self):
        """Test streamed output with tty control codes"""

        stream = TestWrapperIO()

        with \
                patch('aloe.result.Terminal', new=MockTerminal) as mock_term, \
                patch('aloe.result.AloeTestResult.printSummary'):

            mock_term.is_a_tty = True
            self.run_features('-n', '1',
                              'features/highlighting.feature',
                              verbosity=3, stream=stream)

            # we are going to see the scenario written out 3 times
            # once in color 8 as a preview, then each line individually
            # followed by a green version of it
            self.assertEqual(stream.getvalue(), """
t.bold_white(Feature: Highlighting)

t.white(As a programmer
  I want to see my scenarios with pretty highlighting
  So my output is easy to read)

t.bold_white(Scenario: behave_as works)t.color8(features/highlighting.feature:13)
t.color8(Given I have a table
    Given I have entered 10 into the calculator
    And I press [+])
<t.move_up><t.move_up><t.move_up>t.color11(Given I have a table)
<t.move_up>t.bold_green(Given I have a table)
t.color11(Given I have entered 10 into the calculator)
<t.move_up>t.bold_green(Given I have entered 10 into the calculator)
t.color11(And I press [+])
<t.move_up>t.bold_green(And I press [+])

""".lstrip())  # noqa

    def test_full_color_output_no_mocks(self):
        """Test full color output with no mocks"""

        self.assert_feature_success('--color',
                                    'features/highlighting.feature',
                                    verbosity=3)



"""
Test step loading.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()

import os

from aloe.testing import (
    FeatureTest,
    in_directory,
)


@in_directory('tests/step_definition_app')
class StepLoadingTest(FeatureTest):
    """
    Test that calculator feature works as expected.
    """

    def test_single_feature(self):
        """
        Test running a single feature.
        """

        self.assert_feature_success('features/single.feature')

    def test_subdirectory_feature(self):
        """
        Test running a feature in a subdirectory.
        """

        self.assert_feature_success(
            'features/subdirectory/another.feature')

    def test_all_features(self):
        """
        Test running all the features without explicitly specifying them.

        Features not in packages (not_a_submodule, python_inside/package)
        should not be run.
        """

        result = self.assert_feature_success()
        self.assertEqual(result.tests_run, [
            os.path.abspath(feature) for feature in (
                'features/single.feature',
                'features/subdirectory/another.feature',
                'submodule/features/third.feature',
            )
        ])



# -*- coding: utf-8 -*-
"""
Registry of steps and hooks called when running scenarios.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin
from builtins import bytes
from builtins import super
from builtins import str
# pylint:enable=redefined-builtin
from future import standard_library
standard_library.install_aliases()

import re
from collections import OrderedDict
from functools import wraps, partial

from aloe.codegen import multi_manager
from aloe.exceptions import (
    NoDefinitionFound,
    StepLoadingError,
)
from aloe.utils import unwrap_function


# What part of the test to hook
HOOK_WHAT = (
    'all',
    'feature',
    'example',
    'step',
)


# When to execute the hook in relation to the test part
HOOK_WHEN = (
    'before',
    'after',
    'around',
)


class PriorityClass(object):
    """
    Priority class constants.
    """

    DISPLAY = -20  # Display callbacks run first/last
    SYSTEM_OUTER = -10  # System callbacks executing before all except DISPLAY
    USER = 0  # User callbacks
    SYSTEM_INNER = 10  # System callbacks executing after others

    # Note that for 'after' and the 'after' part of 'around' callbacks,
    # the order is reversed


class CallbackDict(dict):
    """
    A collection of callbacks for all situations.
    """

    def __init__(self):
        """
        Initialize the callback lists for every kind of situation.
        """
        super().__init__({
            what: {
                when: {}
                for when in HOOK_WHEN
            }
            for what in HOOK_WHAT
        })

    @classmethod
    def _function_id(cls, func):
        """
        A unique identifier of a function to prevent adding the same hook
        twice.

        To support dynamically generated functions, take the variables from
        the function closure into account.
        """

        func = unwrap_function(func)

        return (
            func.__code__.co_filename,
            func.__code__.co_firstlineno,
            # variables in the closure might not be hashable
            tuple(str(c.cell_contents) for c in func.__closure__ or ()),
        )

    # pylint:disable=too-many-arguments
    def append_to(self, what, when, function, name=None, priority=0):
        """
        Add a callback for a particular type of hook.
        """
        if name is None:
            name = self._function_id(function)

        funcs = self[what][when].setdefault(priority, OrderedDict())
        funcs.pop(name, None)
        funcs[name] = function
    # pylint:enable=too-many-arguments

    def clear(self, name=None, priority_class=None):
        """
        Remove matching callbacks.
        If name is given, only remove callbacks with given name.
        If a priority class is given, only remove ones with the given class.
        """
        for what_dict in self.values():
            for when_dict in what_dict.values():
                if priority_class is None:
                    action_values = when_dict.values()
                else:
                    action_values = (
                        value
                        for (pc, _), value
                        in when_dict.items()
                        if pc == priority_class
                    )
                for callback_list in action_values:
                    if name is None:
                        callback_list.clear()
                    else:
                        callback_list.pop(name, None)

    def hook_list(self, what, when):
        """
        Get all the hooks for a certain event, sorted appropriately.
        """
        return tuple(
            func
            for priority in sorted(self[what][when].keys())
            for func in self[what][when][priority].values()
        )

    def wrap(self, what, function, *hook_args, **hook_kwargs):
        """
        Return a function that executes all the callbacks in proper relations
        to the given test part.
        """

        before_hooks = self.hook_list(what, 'before')
        around_hooks = self.hook_list(what, 'around')
        after_hooks = self.hook_list(what, 'after')

        multi_hook = multi_manager(*around_hooks)

        @wraps(function)
        def wrapped(*args, **kwargs):
            """Run all the hooks in proper relations to the event."""
            for before_hook in before_hooks:
                before_hook(*hook_args, **hook_kwargs)

            try:
                with multi_hook(*hook_args, **hook_kwargs):
                    return function(*args, **kwargs)
            finally:
                # 'after' hooks still run after an exception
                for after_hook in reversed(after_hooks):
                    after_hook(*hook_args, **hook_kwargs)

        return wrapped

    def before_after(self, what):
        """
        Return a pair of functions to execute before and after the event.
        """

        before_hooks = self.hook_list(what, 'before')
        around_hooks = self.hook_list(what, 'around')
        after_hooks = self.hook_list(what, 'after')

        multi_hook = multi_manager(*around_hooks)

        # Save in a closure for both functions
        around_hook = [None]

        def before_func(*args, **kwargs):
            """All hooks to be called before the event."""
            for before_hook in before_hooks:
                before_hook(*args, **kwargs)

            around_hook[0] = multi_hook(*args, **kwargs)
            around_hook[0].__enter__()

        def after_func(*args, **kwargs):
            """All hooks to be called after the event."""
            around_hook[0].__exit__(None, None, None)
            around_hook[0] = None

            for after_hook in after_hooks:
                after_hook(*args, **kwargs)

        return before_func, after_func


class StepDict(dict):
    """
    A mapping of step sentences to their definitions.
    """

    def load(self, sentence, func):
        """Add a mapping between a step sentence and a function."""

        step_re = self._assert_is_step(sentence, func)
        self[step_re] = func

        try:
            func.sentence = sentence
            func.unregister = partial(self.unload, step_re)
        except AttributeError:
            # func might have been a bound method, no way to set attributes
            # on that
            pass

        return func

    def unload(self, sentence):
        """Remove a mapping for a given step sentence, if it exists."""
        try:
            del self[sentence]
        except KeyError:
            pass

    def load_func(self, func):
        """Load a step from a function."""
        sentence = self.extract_sentence(func)
        return self.load(sentence, func)

    def load_steps(self, obj):
        """Load steps from an object."""
        exclude = getattr(obj, "exclude", [])
        for attr in dir(obj):
            if self._attr_is_step(attr, obj) and attr not in exclude:
                step_method = getattr(obj, attr)
                self.load_func(step_method)
        return obj

    def extract_sentence(self, func):
        """Extract the step sentence from a function."""
        func = getattr(func, '__func__', func)
        sentence = getattr(func, '__doc__', None)
        if sentence is None:
            sentence = func.__name__.replace('_', ' ')
            sentence = sentence[0].upper() + sentence[1:]
        return sentence

    def _assert_is_step(self, sentence, func):
        """Compile a step definition or raise an error."""
        try:
            return re.compile(sentence, re.I | re.U)
        except re.error as exc:
            raise StepLoadingError("Error when trying to compile:\n"
                                   "  regex: %r\n"
                                   "  for function: %s\n"
                                   "  error: %s" % (sentence, func, exc))

    def _attr_is_step(self, attr, obj):
        """Test whether an object's attribute is a step."""
        return attr[0] != '_' and self._is_func_or_method(getattr(obj, attr))

    def _is_func_or_method(self, func):
        """Test whether an object is a function or a method."""
        func_dir = dir(func)
        return callable(func) and (
            'func_name' in func_dir or
            '__func__' in func_dir)

    def match_step(self, step_):
        """
        Find a function and arguments to call for a specified Step.

        Returns a tuple of (function, args, kwargs).
        """

        for regex, func in self.items():
            matched = regex.search(step_.sentence)
            if matched:
                kwargs = matched.groupdict()
                if kwargs:
                    return (func, (), matched.groupdict())
                else:
                    args = matched.groups()
                    return (func, args, {})

        raise NoDefinitionFound(step_)

    def step(self, step_func_or_sentence):
        """
        Decorates a function, so that it will become a new step
        definition.

        You give step sentence either (by priority):

         * with step function argument;
         * with function doc; or
         * with the function name exploded by underscores.

        Parameters can be passed to steps using regular expressions.
        Parameters are passed in the order they are captured. Be aware that
        captured values are strings.

        The first parameter passed into the decorated function is the
        :class:`Step` object built for this step.

        Examples:

        .. code-block:: python

            @step("I go to the shops")
            def _i_go_to_the_shops_step(self):
                '''Implements I go to the shops'''

                ...

            @step
            def _i_go_to_the_shops_step(self):
                '''I go to the shops'''

                ...

            @step(r"I buy (\\d+) oranges")
            def _purchase_oranges_step(self, num_oranges):
                '''Buy a certain number of oranges'''

                num_oranges = int(num_oranges)

                ...
        """

        if isinstance(step_func_or_sentence, bytes):
            # Python 2 strings, convert to str
            step_func_or_sentence = step_func_or_sentence.decode()

        if isinstance(step_func_or_sentence, str):
            return lambda func: self.load(step_func_or_sentence, func)
        else:
            return self.load_func(step_func_or_sentence)


STEP_REGISTRY = StepDict()


# This is a function, not a constant
# pylint:disable=invalid-name
step = STEP_REGISTRY.step
# pylint:enable=invalid-name


CALLBACK_REGISTRY = CallbackDict()


class CallbackDecorator(object):
    """
    Add functions to the appropriate callback lists.
    """

    def __init__(self, registry, when, priority_class=PriorityClass.USER):
        self.registry = registry
        self.when = when
        self.priority_class = priority_class

    def _decorate(self, what, function, name=None, priority=0):
        """
        Add the specified function (with name if given) to the callback list.
        """

        priority = (self.priority_class, priority)

        self.registry.append_to(what, self.when, function,
                                name=name,
                                priority=priority)
        return function

    def make_decorator(what):  # pylint:disable=no-self-argument
        """
        Make a decorator for a specific situation.

        NOTE: This is not a method of this class, just used to generate
        methods.
        """

        def decorator(self, function, **kwargs):
            """Decorator method for a particular situation."""
            # pylint:disable=protected-access
            return self._decorate(what, function, **kwargs)
        return decorator

    each_step = make_decorator('step')
    each_example = make_decorator('example')
    each_feature = make_decorator('feature')
    all = make_decorator('all')


# These are functions, not constants
# pylint:disable=invalid-name
after = CallbackDecorator(CALLBACK_REGISTRY, 'after')
around = CallbackDecorator(CALLBACK_REGISTRY, 'around')
before = CallbackDecorator(CALLBACK_REGISTRY, 'before')
# pylint:enable=invalid-name



"""
Utilities for testing libraries using Aloe.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
# pylint:disable=redefined-builtin
from builtins import super
# pylint:enable=redefined-builtin
standard_library.install_aliases()

import io
import os
import sys
import tempfile
import unittest
from contextlib import contextmanager
from functools import wraps

from aloe import world
from aloe.fs import path_to_module_name
from aloe.plugin import GherkinPlugin
from aloe.registry import (
    CALLBACK_REGISTRY,
    PriorityClass,
    STEP_REGISTRY,
)
from aloe.runner import Runner
from aloe.utils import PY3, TestWrapperIO

# When the outer Nose captures output, it's a different type between Python 2
# and 3.
if PY3:
    CAPTURED_OUTPUTS = (io.StringIO,)
else:
    # io.StringIO is an alias to cStringIO.StringIO, which is a function and
    # not a type
    import StringIO  # pylint:disable=import-error
    CAPTURED_OUTPUTS = (
        type(io.StringIO()),
        StringIO.StringIO,
    )


@contextmanager
def _in_directory(directory):
    """
    A context manager to run the payload in a specified directory, with it
    added to the module search path and all modules from it removed from
    sys.modules upon exit.
    """

    directory = os.path.abspath(directory)
    last_wd = os.getcwd()

    sys.path.insert(0, directory)
    os.chdir(directory)

    try:
        yield
    finally:
        os.chdir(last_wd)
        sys.path.remove(directory)

        # Unload modules which are loaded from the directory
        unload_modules = []
        unload_path_prefix = os.path.join(directory, '')
        for module_name, module in sys.modules.items():
            # Find out where is the module loaded from
            try:
                path = module.__file__
            except AttributeError:
                # Maybe a namespace package module?
                try:
                    if module.__spec__.origin == 'namespace':
                        # pylint:disable=protected-access
                        path = module.__path__._path[0]
                        # pylint:enable=protected-access
                    else:
                        continue
                except AttributeError:
                    continue

            # Is it loaded from a file in the directory?
            path = os.path.abspath(path)
            if not path.startswith(unload_path_prefix):
                continue

            # Does its name match what would it be if it was? Consider two
            # directories, foo/ and foo/bar on sys.path, foo/bar/baz.py
            # might have been loaded from either.
            path = path[len(unload_path_prefix):]
            if module_name == path_to_module_name(path):
                unload_modules.append(module_name)

        for module in unload_modules:
            del sys.modules[module]


def in_directory(directory):
    """
    A decorator to change the current directory and add the new one to the
    Python module search path.

    Upon exiting, the directory is changed back, the module search path change
    is reversed and all modules loaded from the directory are removed from
    sys.modules.

    Applies to either a function or an instance of
    TestCase, in which case setUp/tearDown are used.
    """

    def wrapper(func_or_class):
        """
        Wrap a function or a test case class to execute in a different
        directory.
        """

        try:
            is_test_case = issubclass(func_or_class, unittest.TestCase)
        except TypeError:
            is_test_case = False

        if is_test_case:
            # Wrap setUp/tearDown
            old_setup = func_or_class.setUp
            old_teardown = func_or_class.tearDown

            in_directory_cm = [None]

            @wraps(old_setup)
            def setUp(self):
                """Wrap setUp to change to given directory first."""
                in_directory_cm[0] = _in_directory(directory)
                in_directory_cm[0].__enter__()
                old_setup(self)

            @wraps(old_teardown)
            def tearDown(self):
                """Wrap tearDown to restore the original directory."""
                old_teardown(self)
                in_directory_cm[0].__exit__(None, None, None)
                in_directory_cm[0] = None

            func_or_class.setUp = setUp
            func_or_class.tearDown = tearDown

            return func_or_class

        else:
            # Wrap a function
            @wraps(func_or_class)
            def wrapped(*args, **kwargs):
                """
                Execute the function in a different directory.
                """

                with _in_directory(directory):
                    return func_or_class(*args, **kwargs)

            return wrapped

    return wrapper


class TestGherkinPlugin(GherkinPlugin):
    """
    Test Gherkin plugin.
    """

    def loadTestsFromFile(self, file_):
        """
        Record which tests were run.
        """

        for scenario in super().loadTestsFromFile(file_):
            yield scenario

        self.runner.tests_run.append(file_)


class TestRunner(Runner):
    """
    A test test runner to store information about the tests run.

    :param stream: a stream to write the output into (optional)
    """

    def gherkin_plugin(self):
        """
        Override the plugin class.
        """

        plugin = TestGherkinPlugin()
        plugin.runner = self
        return plugin

    def __init__(self, *args, **kwargs):
        self.tests_run = []
        self.stream = kwargs.pop('stream')

        super().__init__(*args, **kwargs)

    def makeConfig(self, *args, **kwargs):
        config = super().makeConfig(*args, **kwargs)

        if self.stream:
            config.stream = self.stream

        return config


class FeatureTest(unittest.TestCase):
    """
    Base class for tests running Gherkin features.
    """

    def setUp(self):
        """
        Ensure inner Nose doesn't redirect output.
        """

        os.environ['NOSE_NOCAPTURE'] = '1'

    def run_feature_string(self, feature_string):
        """
        Run the specified string as a feature.

        The feature will be created as a temporary file in the 'features'
        directory relative to the current directory. This ensures the steps
        contained within would be found by the loader.
        """

        if not os.path.isdir('features'):
            raise ValueError(
                "Features directory not found in {0}".format(os.getcwd()))

        with tempfile.NamedTemporaryFile(suffix='.feature', dir='features') \
                as feature_file:
            feature_file.write(feature_string.encode())
            feature_file.flush()
            return self.run_features(os.path.relpath(feature_file.name))

    def run_features(self, *features, **kwargs):
        """
        Run the specified features.
        """

        # named keyword args and variable positional args aren't supported on
        # Python 2
        verbosity = kwargs.get('verbosity')
        stream = kwargs.get('stream')
        force_color = kwargs.get('force_color', False)

        if stream is None and isinstance(sys.stdout, CAPTURED_OUTPUTS):
            # Don't show results of running the inner tests if the outer Nose
            # redirects output
            stream = TestWrapperIO()

        CALLBACK_REGISTRY.clear(priority_class=PriorityClass.USER)
        STEP_REGISTRY.clear()
        world.__dict__.clear()

        argv = ['aloe']

        if verbosity:
            argv += ['--verbosity', str(verbosity)]

        if force_color:
            argv += ['--color']

        argv += list(features)

        result = TestRunner(exit=False, argv=argv, stream=stream)
        result.captured_stream = stream
        return result

    def assert_feature_success(self, *features, **kwargs):
        """
        Assert that the specified features can be run successfully.
        """

        result = self.run_features(*features, **kwargs)
        try:
            assert result.success
            return result
        except AssertionError:
            if isinstance(result.captured_stream, CAPTURED_OUTPUTS):
                print("--Output--")
                print(result.captured_stream.getvalue())
                print("--END--")
            raise

    def assert_feature_fail(self, *features, **kwargs):
        """
        Assert that the specified features fail when run.
        """

        result = self.run_features(*features, **kwargs)
        try:
            assert not result.success
            return result
        except AssertionError:
            if isinstance(result.captured_stream, CAPTURED_OUTPUTS):
                print("--Output--")
                print(result.captured_stream.getvalue())
                print("--END--")
            raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Main module.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()

import sys
import threading

from aloe.registry import (
    after,
    around,
    before,
    step,
)

world = threading.local()  # pylint:disable=invalid-name


def main(argv=None):  # pragma: no cover
    """
    Entry point for running Aloe.
    """

    if argv is None:
        argv = sys.argv

    from aloe.runner import Runner
    Runner(argv)

if __name__ == '__main__':  # pragma: no cover
    main()



# -*- coding: utf-8 -*-
"""
Test Gherkin parser.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin
from builtins import zip
# pylint:enable=redefined-builtin
from future import standard_library
standard_library.install_aliases()

from nose.tools import assert_equal, assert_raises

from aloe.parser import Feature, Scenario, Background
from aloe.exceptions import LettuceSyntaxError

FEATURE1 = """
Feature: Rent movies
    Scenario: Renting a featured movie
        Given I have the following movies in my database
           | Name                    | Rating  | New | Available |
           | Matrix Revolutions      | 4 stars | no  | 6         |
           | Iron Man 2              | 5 stars | yes | 11        |
        When the client 'John Doe' rents 'Iron man 2'
        Then he needs to pay 10 bucks

    Scenario: Renting a non-featured movie
        Given I have the following movies in my database
           | Name                    | Rating  | New | Available |
           | A night at the museum 2 | 3 stars | yes | 9         |
           | Matrix Revolutions      | 4 stars | no  | 6         |
        When the client 'Mary Doe' rents 'Matrix Revolutions'
        Then she needs to pay 6 bucks

    Scenario: Renting two movies allows client to take one more without charge
        Given I have the following movies in my database
           | Name                    | Rating  | New | Available |
           | A night at the museum 2 | 3 stars | yes | 9         |
           | Matrix Revolutions      | 4 stars | no  | 6         |
           | Iron Man 2              | 5 stars | yes | 11        |
        When the client 'Jack' rents 'Iron man 2'
        And also rents 'Iron man 2' and 'A night at the museum 2'
        Then he needs to pay 16 bucks
"""

FEATURE2 = """
Feature: Division
      In order to avoid silly mistakes
      Cashiers must be able to calculate a fraction

      Scenario: Regular numbers
            Given I have entered 3 into the calculator
            And I have entered 2 into the calculator
            When I press divide
            Then the result should be 1.5 on the screen
"""

FEATURE3 = """
Feature: A long line as feature name will define the max length of the feature
  In order to describe my features
  I want to add description on them
  Scenario: Regular numbers
    Given nothing to do
"""

FEATURE4 = """
Feature: Big sentence
  As a clever guy
  I want to describe this Feature
  So that I can take care of my Scenario
  Scenario: Regular numbers
    Given a huge sentence, that have so many characters
    And another one, very tiny
"""

FEATURE5 = """
Feature: Big table
  Scenario: Regular numbers
    Given that I have these items:
      | description                                                               |
      | this is such a huge description within a table, the maxlengh will be huge |
      | this is another description within a table                                |
"""  # noqa

FEATURE6 = """
Feature: Big scenario outline
  Scenario Outline: Regular numbers
    Given I do fill 'description' with '<value_two>'
  Examples:
    | value_two                                                               |
    | this is such a huge value within a table, the maxlengh will be damn big |
    | this is another description within a table                              |

"""

FEATURE7 = """
Feature: Big table
  Scenario: Regular numbers
    Given that I have these items:
      | description-long-as-hell | name-that-will-make-my-max-length-big |
      | one                      | john                                  |
      | two                      | baby                                  |
"""

FEATURE8 = """
Feature: Big scenario outline
  Scenario Outline: big scenario outlines
    Given I do fill 'description' with '<value_two>'

  Examples:
    | value_two_thousand_and_three | another_one | and_even_bigger |
    | 1                            | um          | one             |
    | 2                            | dois        | two             |
"""

FEATURE9 = """
Feature: Big scenario outline
  Scenario Outline: big scenario outlines
    Given I do fill 'description' with '<value_two>'

  Examples:
    | value_two_thousand_and_three_biiiiiiiiiiiiiiiiiiiiiiiiiiiiig |
    | 1                                                            |
    | 2                                                            |
    | 3                                                            |
"""

FEATURE10 = """
Feature: Big sentence
  As a clever guy
  I want to describe this Feature
  So that I can take care of my Scenario
  Scenario: Regular numbers
    Given a huge sentence, that have so many characters
    And another one, very tiny

    # Feature: Big sentence
    #   As a clever guy
    #   I want to describe this Feature
    #   So that I can take care of my Scenario
    #   Scenario: Regular numbers
    #     Given a huge sentence, that have so many characters
    #     And another one, very tiny
"""

FEATURE11 = """
Feature: Yay tags
  @many @other
  @basic
  @tags @here @:)
  Scenario: Double Yay
    Given this scenario has tags
    Then it can be inspected from within the object
"""

FEATURE12 = """
Feature: Yay tags and many scenarios
  @many @other
  @basic
  @tags @here @:)
  Scenario: Holy tag, Batman
    Given this scenario has tags
    Then it can be inspected from within the object

  @only
  @a-few @tags
  Scenario: Holy guacamole
    Given this scenario has tags
    Then it can be inspected from within the object

"""

FEATURE13 = '''
Feature: correct matching
  @runme
  Scenario: Holy tag, Batman
    Given this scenario has tags
    Then it can be inspected from within the object

  Scenario: This has no tags
    Given this scenario has no tags
    Then I fill my email with gabriel@lettuce.it

  @slow
  Scenario: this is slow
    Given this scenario has tags
    When I fill my email with "gabriel@lettuce.it"
    Then it can be inspected from within the object

  Scenario: Also without tags
    Given this scenario has no tags
    Then I fill my email with 'gabriel@lettuce.it'
'''

FEATURE14 = """
Feature:    Extra whitespace feature
  I want to match scenarios with extra whitespace
  Scenario:    Extra whitespace scenario
    Given this scenario, which has extra leading whitespace
    Then the scenario definition should still match
"""

FEATURE15 = """
Feature: Redis database server

    Scenario: Bootstraping Redis role
        Given I have a an empty running farm
        When I add redis role to this farm
        Then I expect server bootstrapping as M1
        And scalarizr version is last in M1
        And redis is running on M1

    Scenario: Restart scalarizr
        When I reboot scalarizr in M1
        And see 'Scalarizr terminated' in M1 log
        Then scalarizr process is 2 in M1
        And not ERROR in M1 scalarizr log

    @rebundle
    Scenario: Rebundle server
        When I create server snapshot for M1
        Then Bundle task created for M1
        And Bundle task becomes completed for M1

    @rebundle
    Scenario: Use new role
        Given I have a an empty running farm
        When I add to farm role created by last bundle task
        Then I expect server bootstrapping as M1

    @rebundle
    Scenario: Restart scalarizr after bundling
        When I reboot scalarizr in M1
        And see 'Scalarizr terminated' in M1 log
        Then scalarizr process is 2 in M1
        And not ERROR in M1 scalarizr log

    Scenario: Bundling data
        When I trigger databundle creation
        Then Scalr sends DbMsr_CreateDataBundle to M1
        And Scalr receives DbMsr_CreateDataBundleResult from M1
        And Last databundle date updated to current

    Scenario: Modifying data
        Given I have small-sized database 1 on M1
        When I create a databundle
        And I terminate server M1
        Then I expect server bootstrapping as M1
        And M1 contains database 1

    Scenario: Reboot server
        When I reboot server M1
        Then Scalr receives RebootStart from M1
        And Scalr receives RebootFinish from M1

    Scenario: Backuping data on Master
        When I trigger backup creation
        Then Scalr sends DbMsr_CreateBackup to M1
        And Scalr receives DbMsr_CreateBackupResult from M1
        And Last backup date updated to current

    Scenario: Setup replication
        When I increase minimum servers to 2 for redis role
        Then I expect server bootstrapping as M2
        And scalarizr version is last in M2
        And M2 is slave of M1

    Scenario: Restart scalarizr in slave
        When I reboot scalarizr in M2
        And see 'Scalarizr terminated' in M2 log
        Then scalarizr process is 2 in M2
        And not ERROR in M2 scalarizr log

    Scenario: Slave force termination
        When I force terminate M2
        Then Scalr sends HostDown to M1
        And not ERROR in M1 scalarizr log
        And redis is running on M1
        And scalarizr process is 2 in M1
        Then I expect server bootstrapping as M2
        And not ERROR in M1 scalarizr log
        And not ERROR in M2 scalarizr log
        And redis is running on M1

    @ec2
    Scenario: Slave delete EBS
        When I know M2 ebs storage
        And M2 ebs status is in-use
        Then I terminate server M2 with decrease
        And M2 ebs status is deleting
        And not ERROR in M1 scalarizr log

    @ec2
    Scenario: Setup replication for EBS test
        When I increase minimum servers to 2 for redis role
        Then I expect server bootstrapping as M2
        And M2 is slave of M1

    Scenario: Writing on Master, reading on Slave
        When I create database 2 on M1
        Then M2 contains database 2

    Scenario: Slave -> Master promotion
        Given I increase minimum servers to 3 for redis role
        And I expect server bootstrapping as M3
        When I create database 3 on M1
        And I terminate server M1 with decrease
        Then Scalr sends DbMsr_PromoteToMaster to N1
        And Scalr receives DbMsr_PromoteToMasterResult from N1
        And Scalr sends DbMsr_NewMasterUp to all
        And M2 contains database 3

    @restart_farm
    Scenario: Restart farm
        When I stop farm
        And wait all servers are terminated
        Then I start farm
        And I expect server bootstrapping as M1
        And scalarizr version is last in M1
        And redis is running on M1
        And M1 contains database 3
        Then I expect server bootstrapping as M2
        And M2 is slave of M1
        And M2 contains database 3
        """

FEATURE18 = """
@feature_runme
Feature: correct matching
  @runme1
  Scenario: Holy tag, Batman [1]
    Given this scenario has tags
    Then it can be inspected from within the object

  @runme2
  Scenario: Holy tag2, Batman (2)
    Given this scenario has other tags
    Then it can be inspected from within the object even with the table
    | What | Is | This  |
    | It   | is | TABLE |

  @runme3
  Scenario: Holy tag3, Batman
    Given this scenario has even more tags
    Then it can be inspected from within the object

"""


FEATURE19 = """
Feature: correct matching
  @runme1
  Scenario: Holy tag, Batman (1)
    Given this scenario has tags
    Then it can be inspected from within the object

  @runme2
  Scenario: Holy tag2, Batman [2]
    Given this scenario has other tags
    Then it can be inspected from within the object
"""

FEATURE16 = """
Feature: Movie rental
    As a rental store owner
    I want to keep track of my clients
    So that I can manage my business better

    Background:
        Given I have the following movies in my database:
           | Name                    | Rating  | New | Available |
           | Matrix Revolutions      | 4 stars | no  | 6         |
           | Iron Man 2              | 5 stars | yes | 11        |
        And the following clients:
           | Name      |
           | John Doe  |
           | Foo Bar   |

    Scenario: Renting a featured movie
        Given the client 'John Doe' rents 'Iron Man 2'
        Then there are 10 more left

    Scenario: Renting an old movie
        Given the client 'Foo Bar' rents 'Matrix Revolutions'
        Then there are 5 more left
"""

FEATURE17 = """
Feature: Movie rental without MMF
    Background:
        Given I have the following movies in my database:
           | Name                    | Rating  | New | Available |
           | Matrix Revolutions      | 4 stars | no  | 6         |
           | Iron Man 2              | 5 stars | yes | 11        |
        And the following clients:
           | Name      |
           | John Doe  |
           | Foo Bar   |

    Scenario: Renting a featured movie
        Given the client 'John Doe' rents 'Iron Man 2'
        Then there are 10 more left
"""

FEATURE20 = """
Feature: My scenarios have no name
    Scenario:
        Given this scenario raises a syntax error
"""

FEATURE21 = """
Feature: Taming the tag parser

  Background:
    Given the email addresses:
      | name         | email                      |
      | Chuck Norris | roundhouse@chucknorris.com |
    Then the next scenario has only the tags it's supposed to

  Scenario: I'm isolated
    Given I am parsed
    Then this scenario has only zero tags

  @tag
  Scenario: I'm so isolated
    Given I am parsed
    Then this scenario has one tag
"""

FEATURE22 = """
Feature: one tag in the first scenario

  @onetag
  Scenario: This is the first scenario
    Given I am parsed
    Then this scenario has one tag
"""

FEATURE23 = """
Feature: three tags in the first scenario

  @onetag @another @$%^&even-weird_chars
  Scenario: This is the first scenario
    Given I am parsed
    Then this scenario has three tags
"""


def test_feature_has_repr():
    """
    Feature implements __repr__ nicely
    """
    feature = Feature.from_string(FEATURE1)
    assert repr(feature) == '<Feature: "Rent movies">'


def test_scenario_has_name():
    """
    It should extract the name string from the scenario
    """

    feature = Feature.from_string(FEATURE1)

    assert isinstance(feature, Feature)

    assert feature.name == "Rent movies"


def test_feature_has_scenarios():
    """
    A feature object should have a list of scenarios
    """

    feature = Feature.from_string(FEATURE1)

    assert isinstance(feature.scenarios, tuple)
    assert len(feature.scenarios) == 3

    expected_names = [
        "Renting a featured movie",
        "Renting a non-featured movie",
        "Renting two movies allows client to take one more without charge",
    ]

    for scenario, expected_name in zip(feature.scenarios, expected_names):
        assert isinstance(scenario, Scenario)
        assert scenario.name == expected_name

    assert_equal(
        feature.scenarios[1].steps[0].keys,
        ('Name', 'Rating', 'New', 'Available')
    )

    assert list(feature.scenarios[1].steps[0].hashes) == [
        {
            'Name': 'A night at the museum 2',
            'Rating': '3 stars',
            'New': 'yes',
            'Available': '9',
        },
        {
            'Name': 'Matrix Revolutions',
            'Rating': '4 stars',
            'New': 'no',
            'Available': '6',
        },
    ]


def test_can_parse_feature_description():
    """
    A feature object should have a description
    """

    feature = Feature.from_string(FEATURE2)

    assert_equal(
        feature.description,
        "In order to avoid silly mistakes\n"
        "Cashiers must be able to calculate a fraction"
    )
    expected_scenario_names = ["Regular numbers"]
    got_scenario_names = [s.name for s in feature.scenarios]

    assert_equal(expected_scenario_names, got_scenario_names)
    assert_equal(len(feature.scenarios[0].steps), 4)

    step1, step2, step3, step4 = feature.scenarios[0].steps

    assert_equal(step1.sentence, 'Given I have entered 3 into the calculator')
    assert_equal(step2.sentence, 'And I have entered 2 into the calculator')
    assert_equal(step3.sentence, 'When I press divide')
    assert_equal(step4.sentence,
                 'Then the result should be 1.5 on the screen')


def test_scenarios_parsed_by_feature_has_feature():
    "Scenarios parsed by features has feature"

    feature = Feature.from_string(FEATURE2)

    for scenario in feature.scenarios:
        assert_equal(scenario.feature, feature)


def test_feature_max_length_on_scenario():
    """
    The max length of a feature considering when the scenario is longer than
    the remaining things
    """

    feature = Feature.from_string(FEATURE1)
    assert_equal(feature.max_length, 76)


def test_feature_max_length_on_feature_description():
    """
    The max length of a feature considering when one of the description lines
    of the feature is longer than the remaining things
    """

    feature = Feature.from_string(FEATURE2)
    assert_equal(feature.max_length, 47)


def test_feature_max_length_on_feature_name():
    """
    The max length of a feature considering when the name of the feature
    is longer than the remaining things
    """

    feature = Feature.from_string(FEATURE3)
    assert_equal(feature.max_length, 78)


def test_feature_max_length_on_step_sentence():
    """
    The max length of a feature considering when the some of the step sentences
    is longer than the remaining things
    """

    feature = Feature.from_string(FEATURE4)
    assert_equal(feature.max_length, 55)


def test_feature_max_length_on_step_with_table():
    """
    The max length of a feature considering when the table of some of the steps
    is longer than the remaining things
    """

    feature = Feature.from_string(FEATURE5)
    assert_equal(feature.max_length, 83)


def test_feature_max_length_on_step_with_table_keys():
    """
    The max length of a feature considering when the table keys of some of the
    steps are longer than the remaining things
    """

    feature = Feature.from_string(FEATURE7)
    assert_equal(feature.max_length, 74)


def test_feature_max_length_on_scenario_outline():
    """
    The max length of a feature considering when the table of some of the
    scenario oulines is longer than the remaining things
    """

    feature = Feature.from_string(FEATURE6)
    assert_equal(feature.max_length, 79)


def test_feature_max_length_on_scenario_outline_keys():
    """
    The max length of a feature considering when the table keys of the
    scenario oulines are longer than the remaining things
    """

    feature1 = Feature.from_string(FEATURE8)
    feature2 = Feature.from_string(FEATURE9)
    assert_equal(feature1.max_length, 68)
    assert_equal(feature2.max_length, 68)


def test_description_on_long_named_feature():
    "Can parse the description on long named features"
    feature = Feature.from_string(FEATURE3)
    assert_equal(
        feature.description,
        "In order to describe my features\n"
        "I want to add description on them",
    )


def test_description_on_big_sentenced_steps():
    "Can parse the description on long sentenced steps"
    feature = Feature.from_string(FEATURE4)
    assert_equal(
        feature.description,
        "As a clever guy\n"
        "I want to describe this Feature\n"
        "So that I can take care of my Scenario",
    )


def test_comments():
    """
    It should ignore lines that start with #, despite white spaces
    """

    feature = Feature.from_string(FEATURE10)

    assert_equal(feature.max_length, 55)


def test_single_scenario_single_scenario():
    "Features should have at least the first scenario parsed with tags"
    feature = Feature.from_string(FEATURE11)

    first_scenario = feature.scenarios[0]

    assert_equal(first_scenario.tags, (
        'many', 'other', 'basic', 'tags', 'here', ':)'
    ))


def test_single_feature_single_tag():
    "All scenarios within a feature inherit the feature's tags"
    feature = Feature.from_string(FEATURE18)

    assert feature.scenarios[0].tags == ('runme1', 'feature_runme')

    assert feature.scenarios[1].tags == ('runme2', 'feature_runme')

    assert feature.scenarios[2].tags == ('runme3', 'feature_runme')


def test_single_scenario_many_scenarios():
    "Untagged scenario following a tagged one should have no tags"

    feature = Feature.from_string(FEATURE13)

    first_scenario = feature.scenarios[0]
    assert first_scenario.tags == ('runme',)

    second_scenario = feature.scenarios[1]
    assert second_scenario.tags == ()

    third_scenario = feature.scenarios[2]
    assert third_scenario.tags == ('slow',)

    last_scenario = feature.scenarios[3]
    assert last_scenario.tags == ()


def test_scenarios_with_extra_whitespace():
    "Make sure that extra leading whitespace is ignored"
    feature = Feature.from_string(FEATURE14)

    assert_equal(type(feature.scenarios), tuple)
    assert_equal(len(feature.scenarios), 1, "It should have 1 scenario")
    assert_equal(feature.name, "Extra whitespace feature")

    scenario = feature.scenarios[0]
    assert_equal(type(scenario), Scenario)
    assert_equal(scenario.name, "Extra whitespace scenario")


def test_scenarios_parsing():
    "Tags are parsed correctly"
    feature = Feature.from_string(FEATURE15)
    scenarios_and_tags = [(s.name, s.tags) for s in feature.scenarios]

    assert_equal(scenarios_and_tags, [
        ('Bootstraping Redis role', ()),
        ('Restart scalarizr', ()),
        ('Rebundle server', ('rebundle',)),
        ('Use new role', ('rebundle',)),
        ('Restart scalarizr after bundling', ('rebundle',)),
        ('Bundling data', ()),
        ('Modifying data', ()),
        ('Reboot server', ()),
        ('Backuping data on Master', ()),
        ('Setup replication', ()),
        ('Restart scalarizr in slave', ()),
        ('Slave force termination', ()),
        ('Slave delete EBS', ('ec2',)),
        ('Setup replication for EBS test', ('ec2',)),
        ('Writing on Master, reading on Slave', ()),
        ('Slave -> Master promotion', ()),
        ('Restart farm', ('restart_farm',)),
    ])


def test_scenarios_with_special_characters():
    "Make sure that regex special characters in the scenario names are ignored"
    feature = Feature.from_string(FEATURE19)

    assert feature.scenarios[0].tags == ('runme1',)

    assert feature.scenarios[1].tags == ('runme2',)


def test_background_parsing_with_mmf():
    """Test background parsing with description."""
    feature = Feature.from_string(FEATURE16)
    assert feature.description == \
        "As a rental store owner\n" \
        "I want to keep track of my clients\n" \
        "So that I can manage my business better"

    assert isinstance(feature.background, Background)
    assert feature.background.steps
    assert len(feature.background.steps) == 2

    step1, step2 = feature.background.steps
    assert step1.sentence == \
        'Given I have the following movies in my database:'
    assert_equal(step1.hashes, (
        {
            u'Available': u'6',
            u'Rating': u'4 stars',
            u'Name': u'Matrix Revolutions',
            u'New': u'no',
        },
        {
            u'Available': u'11',
            u'Rating': u'5 stars',
            u'Name': u'Iron Man 2',
            u'New': u'yes',
        },
    ))

    assert step2.sentence == \
        'And the following clients:'
    assert_equal(step2.hashes, (
        {u'Name': u'John Doe'},
        {u'Name': u'Foo Bar'},
    ))


def test_background_parsing_without_mmf():
    """Test background parsing without description."""
    feature = Feature.from_string(FEATURE17)
    assert feature.description == ""

    assert isinstance(feature.background, Background)
    assert feature.background.steps
    assert len(feature.background.steps) == 2

    step1, step2 = feature.background.steps
    assert step1.sentence == \
        'Given I have the following movies in my database:'
    assert_equal(step1.hashes, (
        {
            u'Available': u'6',
            u'Rating': u'4 stars',
            u'Name': u'Matrix Revolutions',
            u'New': u'no',
        },
        {
            u'Available': u'11',
            u'Rating': u'5 stars',
            u'Name': u'Iron Man 2',
            u'New': u'yes',
        },
    ))
    assert_equal(step1.table, (
        ('Name', 'Rating', 'New', 'Available'),
        ('Matrix Revolutions', '4 stars', 'no', '6'),
        ('Iron Man 2', '5 stars', 'yes', '11'),
    ))

    assert step2.sentence == \
        'And the following clients:'
    assert_equal(step2.hashes, (
        {u'Name': u'John Doe'},
        {u'Name': u'Foo Bar'},
    ))
    assert_equal(step2.table, (
        ('Name',),
        ('John Doe',),
        ('Foo Bar',),
    ))


def test_syntax_error_for_scenarios_with_no_name():
    ("Trying to parse features with unnamed "
     "scenarios will cause a syntax error")
    with assert_raises(LettuceSyntaxError) as error:
        Feature.from_string(FEATURE20)

    assert error.exception.msg == \
        'Syntax error at: None\n' \
        '3:5 Scenario must have a name'


def test_scenario_post_email():
    ("Having a scenario which the body has an email address; "
     "Then the following scenario should have no "
     "tags related to the email")

    feature = Feature.from_string(FEATURE21)
    scenario1, scenario2 = feature.scenarios

    assert scenario1.tags == ()
    assert scenario2.tags == ('tag',)


def test_feature_first_scenario_tag_extraction():
    ("A feature object should be able to find the single tag "
     "belonging to the first scenario")
    feature = Feature.from_string(FEATURE22)

    assert feature.scenarios[0].tags == ('onetag',)


def test_feature_first_scenario_tags_extraction():
    ("A feature object should be able to find the tags "
     "belonging to the first scenario")
    feature = Feature.from_string(FEATURE23)

    assert feature.scenarios[0].tags == \
        ('onetag', 'another', '$%^&even-weird_chars')



"""
Extensions to Sphinx_ for documenting Aloe packages.

Add these extensions to your Sphinx_ ``conf.py``:

.. code-block:: python

    extensions = [
        'sphinx.ext.autodoc',
        'aloe_sphinx.gerkindomain',
        'aloe_sphinx.autosteps',
    ]

Gherkin Domain
~~~~~~~~~~~~~~

.. object:: aloe_sphinx.gherkindomain

.. automodule:: aloe_sphinx.gherkindomain

Steps Autodocumenter
~~~~~~~~~~~~~~~~~~~~

.. object:: aloe_sphinx.autosteps

.. automodule:: aloe_sphinx.autosteps

.. _Sphinx: http://sphinx-doc.org/
"""



"""
Exception classes
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin
from builtins import super
# pylint:enable=redefined-builtin
from future import standard_library
standard_library.install_aliases()


class LettuceSyntaxError(SyntaxError):
    """A syntax error in a feature file."""
    def __init__(self, filename, string):
        self.filename = filename
        self.string = string

        msg = "Syntax error at: {filename}\n{string}".format(
            filename=filename, string=string)

        super().__init__(msg)


class LettuceSyntaxWarning(SyntaxWarning):
    """A warning about syntax in a feature file."""
    pass


class StepLoadingError(Exception):
    """Raised when a step cannot be loaded."""
    pass


class NoDefinitionFound(Exception):
    """
    Exception raised when there is no suitable step definition for a step.
    """

    def __init__(self, step):
        self.step = step
        super().__init__('The step r"%s" is not defined' % self.step.sentence)



"""
Code generation helpers
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()

import ast
import re
from contextlib import contextmanager

from aloe.utils import identifier


FUNCTION_DEF_SAMPLE = ast.parse('def func(): pass')


def make_function(source, context=None, source_file=None, name=None):
    """
    Compile and evaluate given source to a function given the specified
    globals.
    Optionally set the file and name of the function.
    """

    func = ast.parse(source)

    # Check that generated code is a function
    # pylint:disable=unidiomatic-typecheck
    if type(func) != type(FUNCTION_DEF_SAMPLE) \
            or len(func.body) != 1 \
            or type(func.body[0]) != type(FUNCTION_DEF_SAMPLE.body[0]):
        raise ValueError("source must be a function definition.")
    # pylint:enable=unidiomatic-typecheck

    # Set or record the function name
    if name is not None:
        func.body[0].name = name = identifier(name)
    else:
        name = func.body[0].name

    # TODO: What's a better default for file?
    if source_file is None:
        source_file = '<generated>'

    context = context or {}

    code = compile(func, source_file, 'exec')
    eval(code, context)  # pylint:disable=eval-used

    return context[name]


def indent(source, count=1):
    """
    Indent the source by count*4 spaces.
    """

    prepend = ' ' * 4 * count
    return '\n'.join(
        # Only indent the non-empty lines
        prepend + line if line else line
        for line in source.split('\n')
    )


NOT_SPACE = re.compile('[^ ]')


def remove_indent(source):
    """
    Remove as much indentation as possible from the code.
    """

    lines = source.split('\n')
    min_indent = min(
        match.start() for match in (
            NOT_SPACE.search(line)
            for line in lines
        ) if match
    )

    return '\n'.join(
        line[min_indent:]
        for line in lines
    )


def multi_manager(*managers):
    """
    A context manager invoking all the given context managers in order.

    Returns a tuple with all the manager results.
    """

    if len(managers) == 0:
        source = remove_indent(
            """
            def null_manager(*args, **kwargs):
                yield ()
            """
        )
    else:
        with_stmt = ', '.join(
            "manager{i}(*args, **kwargs) as result{i}".format(i=i)
            for i in range(len(managers))
        )

        result_tuple = '(' + ', '.join(
            "result{i}".format(i=i)
            for i in range(len(managers))
        ) + ')'

        source = remove_indent(
            """
            def multi_manager(*args, **kwargs):
                with {with_stmt}:
                    yield {result_tuple}
            """
        ).format(with_stmt=with_stmt, result_tuple=result_tuple)

    context = {
        'manager' + str(i): manager
        for i, manager in enumerate(managers)
    }

    return contextmanager(make_function(
        source=source,
        context=context,
    ))



# -*- coding: utf-8 -*-
"""
Test Portuguese (Brasilian) language parsing.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin,wildcard-import,unused-wildcard-import
from builtins import *
# pylint:enable=redefined-builtin,wildcard-import,unused-wildcard-import
from future import standard_library
standard_library.install_aliases()

from nose.tools import assert_equal

from aloe.parser import Feature

SCENARIO = u"""
Cenário: Consolidar o banco de dados de cursos universitários em arquivo texto
    Dados os seguintes cursos cadastrados no banco de dados da universidade:
       | Nome                    | Duração  |
       | Ciência da Computação   | 5 anos   |
       | Nutrição                | 4 anos   |
    Quando eu consolido os dados no arquivo 'cursos.txt'
    Então a 1a linha do arquivo 'cursos.txt' contém 'Ciência da Computação:5'
    E a 2a linha do arquivo 'cursos.txt' contém 'Nutrição:4'
"""

SCENARIO_OUTLINE1 = u'''
Esquema do Cenário: Cadastrar um aluno no banco de dados
    Dado que eu preencho o campo "nome" com "<nome>"
    E que eu preencho o campo "idade" com "<idade>"
    Quando eu salvo o formulário
    Então vejo a mensagem "Aluno <nome>, de <idade> anos foi cadastrado com sucesso!"

Exemplos:
    | nome    | idade |
    | Gabriel | 22    |
    | João    | 30    |
'''  # noqa

SCENARIO_OUTLINE2 = u'''
Esquema do Cenário: Cadastrar um aluno no banco de dados
    Dado que eu preencho o campo "nome" com "<nome>"
    E que eu preencho o campo "idade" com "<idade>"
    Quando eu salvo o formulário
    Então vejo a mensagem "Aluno <nome>, de <idade> anos foi cadastrado com sucesso!"

Cenários:
    | nome    | idade |
    | Gabriel | 99    |
    | João    | 100   |
'''  # noqa

FEATURE = u'''
Funcionalidade: Pesquisar alunos com matrícula vencida
  Como gerente financeiro
  Eu quero pesquisar alunos com matrícula vencida
  Para propor um financiamento

  Cenário: Pesquisar por nome do curso
    Dado que eu preencho o campo "nome do curso" com "Nutrição"
    Quando eu clico em "pesquisar"
    Então vejo os resultados:
      | nome  | valor devido |
      | João  | R$ 512,66    |
      | Maria | R$ 998,41    |
      | Ana   | R$ 231,00    |
'''


def parse_scenario(string, language=None):
    """Parse a scenario, prefixing it with a feature header."""
    feature = """
Funcionalidade: parse_scenario
    """

    feature += string
    feature = Feature.from_string(feature, language=language)

    return feature.scenarios[0]


def test_scenario_ptbr_from_string():
    """
    Language: PT-BR -> Scenario.from_string
    """

    scenario = parse_scenario(SCENARIO, language='pt-br')

    assert_equal(
        scenario.name,
        u'Consolidar o banco de dados de cursos universitários em '
        u'arquivo texto'
    )
    assert_equal(
        scenario.steps[0].hashes,
        (
            {'Nome': u'Ciência da Computação', u'Duração': '5 anos'},
            {'Nome': u'Nutrição', u'Duração': '4 anos'},
        )
    )


def test_scenario_outline1_ptbr_from_string():
    """
    Language: PT-BR -> Scenario.from_string, with scenario outline, first case
    """

    scenario = parse_scenario(SCENARIO_OUTLINE1, language='pt-br')

    assert_equal(
        scenario.name,
        'Cadastrar um aluno no banco de dados'
    )
    assert_equal(
        scenario.outlines,
        (
            {'nome': u'Gabriel', u'idade': '22'},
            {'nome': u'João', u'idade': '30'},
        )
    )


def test_scenario_outline2_ptbr_from_string():
    """
    Language: PT-BR -> Scenario.from_string, with scenario outline, second case
    """

    scenario = parse_scenario(SCENARIO_OUTLINE2, language='pt-br')

    assert_equal(
        scenario.name,
        'Cadastrar um aluno no banco de dados'
    )
    assert_equal(
        scenario.outlines,
        (
            {'nome': u'Gabriel', u'idade': '99'},
            {'nome': u'João', u'idade': '100'},
        )
    )


def test_feature_ptbr_from_string():
    """
    Language: PT-BR -> Feature.from_string
    """

    feature = Feature.from_string(FEATURE, language='pt-br')

    assert_equal(
        feature.name,
        u'Pesquisar alunos com matrícula vencida'
    )

    assert_equal(
        feature.description,
        u"Como gerente financeiro\n"
        u"Eu quero pesquisar alunos com matrícula vencida\n"
        u"Para propor um financiamento"
    )

    (scenario, ) = feature.scenarios

    assert_equal(
        scenario.name,
        'Pesquisar por nome do curso'
    )

    assert_equal(
        scenario.steps[-1].hashes,
        (
            {'nome': u'João', u'valor devido': 'R$ 512,66'},
            {'nome': u'Maria', u'valor devido': 'R$ 998,41'},
            {'nome': u'Ana', u'valor devido': 'R$ 231,00'},
        )
    )



# -*- coding: utf-8 -*-
"""
Test parsing scenarios.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin
from builtins import zip
# pylint:enable=redefined-builtin
from future import standard_library
standard_library.install_aliases()

from unittest import skip

from aloe.parser import Step, Scenario, Feature
from aloe.exceptions import LettuceSyntaxError

from nose.tools import assert_equal
from nose.tools import assert_raises


SCENARIO1 = """
Scenario: Adding some students to my university database
    Given I have the following courses in my university:
       | Name               | Duration |
       | Computer Science   | 5 years  |
       | Nutrition          | 4 years  |
    When I consolidate the database into 'courses.txt'
    Then I see the 1st line of 'courses.txt' has 'Computer Science:5'
    And I see the 2nd line of 'courses.txt' has 'Nutrition:4'
"""

OUTLINED_SCENARIO = """
Scenario Outline: Add two numbers
    Given I have entered <input_1> into the calculator
    And I have entered <input_2> into the calculator
    When I press <button>
    Then the result should be <output> on the screen

    Examples:
      | input_1 | input_2 | button | output |
      | 20      | 30      | add    | 50     |
      | 2       | 5       | add    | 7      |
      | 0       | 40      | add    | 40     |
"""

OUTLINED_SCENARIO_WITH_SUBSTITUTIONS_IN_TABLE = """
Scenario Outline: Bad configuration should fail
    Given I provide the following configuration:
       | Parameter | Value |
       |     a     |  <a>  |
       |     b     |  <b>  |
    When I run the program
    Then it should fail hard-core

Examples:
    | a | b |
    | 1 | 2 |
    | 2 | 4 |
"""

OUTLINED_SCENARIO_WITH_SUBSTITUTIONS_IN_MULTILINE = '''
Scenario Outline: Parsing HTML
    When I parse the HTML:
        """
        <div><v></div>
        """
    Then I should see "outline value"

Examples:
    | v             |
    | outline value |
'''

OUTLINED_FEATURE = """
    Feature: Do many things at once
        In order to automate tests
        As a automation freaky
        I want to use scenario outlines

        Scenario Outline: Add two numbers wisely
            Given I have entered <input_1> into the calculator
            And I have entered <input_2> into the calculator
            When I press <button>
            Then the result should be <output> on the screen

        Examples:
            | input_1 | input_2 | button | output |
            | 20      | 30      | add    | 50     |
            | 2       | 5       | add    | 7      |
            | 0       | 40      | add    | 40     |
"""

OUTLINED_FEATURE_WITH_MANY = """
    Feature: Full-featured feature
        Scenario Outline: Do something
            Given I have entered <input_1> into the <input_2>

        Examples:
            | input_1 | input_2 |
            | ok      | fail    |
            | fail    | ok      |

        Scenario: Do something else
          Given I am fine

        Scenario: Worked!
          Given it works
          When I look for something
          Then I find it

        Scenario Outline: Add two numbers wisely
            Given I have entered <input_1> into the calculator
            And I have entered <input_2> into the calculator
            When I press <button>
            Then the result should be <output> on the screen

        Examples:
            | input_1 | input_2 | button | output |
            | 20      | 30      | add    | 50     |
            | 2       | 5       | add    | 7      |
            | 0       | 40      | add    | 40     |

        Examples:
            | input_1 | input_2 | button | output |
            | 5       | 7       | add    | 12     |

"""

SCENARIO_FAILED = """
Scenario: Adding some students to my university database
       | Name               | Duration |
       | Computer Science   | 5 years  |
       | Nutrition          | 4 years  |
    When I consolidate the database into 'courses.txt'
    Then I see the 1st line of 'courses.txt' has 'Computer Science:5'
    And I see the 2nd line of 'courses.txt' has 'Nutrition:4'
"""

OUTLINED_SCENARIO_WITH_COMMENTS_ON_EXAMPLES = """
Scenario Outline: Add two numbers
    Given I have entered <input_1> into the calculator
    And I have entered <input_2> into the calculator
    When I press <button>
    Then the result should be <output> on the screen

    Examples:
      | input_1 | input_2 | button | output |
      | 20      | 30      | add    | 50     |
      #| 2       | 5       | add    | 7      |
      | 0       | 40      | add    | 40     |
    # end of the scenario
"""

OUTLINED_SCENARIO_WITH_MORE_THAN_ONE_EXAMPLES_BLOCK = """
Scenario Outline: Add two numbers
    Given I have entered <input_1> into the calculator
    And I have entered <input_2> into the calculator
    When I press <button>
    Then the result should be <output> on the screen

    Examples:
      | input_1 | input_2 | button | output |
      | 20      | 30      | add    | 50     |
      | 2       | 5       | add    | 7      |
      | 0       | 40      | add    | 40     |

    Examples:
      | input_1 | input_2 | button | output |
      | 20      | 33      | add    | 53     |
      | 12      | 40      | add    | 52     |
"""

COMMENTED_SCENARIO = """
Scenario: Adding some students to my university database
    Given I have the following courses in my university:
       | Name               | Duration |
       | Computer Science   | 5 years  |
       | Nutrition          | 4 years  |
    When I consolidate the database into 'courses.txt'
    Then I see the 1st line of 'courses.txt' has 'Computer Science:5'
    And I see the 2nd line of 'courses.txt' has 'Nutrition:4'

# Scenario: Adding some students to my university database
#     Given I have the following courses in my university:
#        | Name               | Duration |
#        | Computer Science   | 5 years  |
#        | Nutrition          | 4 years  |
#     When I consolidate the database into 'courses.txt'
#     Then I see the 1st line of 'courses.txt' has 'Computer Science:5'
#     And I see the 2nd line of 'courses.txt' has 'Nutrition:4'

"""

INLINE_COMMENTS_IGNORED_WITHIN_DOUBLE_QUOTES = '''
Scenario: Tweeting
  Given I am logged in on twitter
  When I search for the hashtag "#hammer"
'''

INLINE_COMMENTS_IGNORED_WITHIN_SINGLE_QUOTES = """
Scenario: Tweeting
  Given I am logged in on twitter
  When I search for the hashtag '#hammer'
"""


def parse_scenario(scenario, tags=None):
    """Parse a scenario, adding a feature header and tags."""
    feature_str = """
    Feature: test scenario
    """

    if tags:
        feature_str += ' '.join('@%s' % tag for tag in tags)

    feature_str += scenario

    feature = Feature.from_string(feature_str)

    return feature.scenarios[0]


def solved_steps(scenario):
    """A list of all the outline steps for a scenario."""

    return tuple(
        step
        for _, steps in scenario.evaluated
        for step in steps
    )


def test_scenario_has_name():
    """
    It should extract the name of the scenario
    """

    scenario = parse_scenario(SCENARIO1)

    assert isinstance(scenario, Scenario)

    assert_equal(
        scenario.name,
        "Adding some students to my university database"
    )


def test_scenario_has_repr():
    """
    Scenario implements __repr__ nicely
    """

    scenario = parse_scenario(SCENARIO1)
    assert_equal(
        repr(scenario),
        '<Scenario: "Adding some students to my university database">'
    )


def test_scenario_has_steps():
    """
    A scenario object should have a list of steps
    """

    scenario = parse_scenario(SCENARIO1)

    assert_equal(type(scenario.steps), tuple)
    assert_equal(len(scenario.steps), 4, "It should have 4 steps")

    expected_sentences = [
        "Given I have the following courses in my university:",
        "When I consolidate the database into 'courses.txt'",
        "Then I see the 1st line of 'courses.txt' has 'Computer Science:5'",
        "And I see the 2nd line of 'courses.txt' has 'Nutrition:4'",
    ]

    for step, expected_sentence in zip(scenario.steps, expected_sentences):
        assert_equal(type(step), Step)
        assert_equal(step.sentence, expected_sentence)

    assert_equal(scenario.steps[0].keys, ('Name', 'Duration'))
    assert_equal(
        scenario.steps[0].hashes,
        (
            {'Name': 'Computer Science', 'Duration': '5 years'},
            {'Name': 'Nutrition', 'Duration': '4 years'},
        )
    )


def test_scenario_may_own_outlines():
    """
    A scenario may own outlines
    """

    scenario = parse_scenario(OUTLINED_SCENARIO)

    assert_equal(len(scenario.steps), 4)
    expected_sentences = [
        'Given I have entered <input_1> into the calculator',
        'And I have entered <input_2> into the calculator',
        'When I press <button>',
        'Then the result should be <output> on the screen',
    ]

    for step, expected_sentence in zip(scenario.steps, expected_sentences):
        assert_equal(type(step), Step)
        assert_equal(step.sentence, expected_sentence)

    assert_equal(scenario.name, "Add two numbers")
    assert_equal(
        scenario.outlines,
        (
            {'input_1': '20', 'input_2': '30',
             'button': 'add', 'output': '50'},
            {'input_1': '2', 'input_2': '5',
             'button': 'add', 'output': '7'},
            {'input_1': '0', 'input_2': '40',
             'button': 'add', 'output': '40'},
        )
    )


def test_steps_parsed_by_scenarios_has_scenarios():
    """
    Steps parsed by scenarios has scenarios
    """

    scenario = parse_scenario(SCENARIO1)
    for step in scenario.steps:
        assert_equal(step.scenario, scenario)


def test_scenario_sentences_can_be_solved():
    """
    A scenario with outlines may solve its sentences
    """
    scenario = parse_scenario(OUTLINED_SCENARIO)
    solved = solved_steps(scenario)

    assert_equal(len(solved), 12)
    expected_sentences = [
        'Given I have entered 20 into the calculator',
        'And I have entered 30 into the calculator',
        'When I press add',
        'Then the result should be 50 on the screen',
        'Given I have entered 2 into the calculator',
        'And I have entered 5 into the calculator',
        'When I press add',
        'Then the result should be 7 on the screen',
        'Given I have entered 0 into the calculator',
        'And I have entered 40 into the calculator',
        'When I press add',
        'Then the result should be 40 on the screen',
    ]

    for step, expected in zip(solved, expected_sentences):
        assert_equal(type(step), Step)
        assert_equal(step.sentence, expected)


def test_scenario_tables_are_solved_against_outlines():
    """
    Outline substitution should apply to tables within a scenario
    """

    expected_hashes_per_step = [
        # a = 1, b = 2
        ({'Parameter': 'a', 'Value': '1'},
         {'Parameter': 'b', 'Value': '2'}),  # Given ...
        (),  # When I run the program
        (),  # Then I crash hard-core

        # a = 2, b = 4
        ({'Parameter': 'a', 'Value': '2'},
         {'Parameter': 'b', 'Value': '4'}),
        (),
        ()
    ]

    scenario = parse_scenario(OUTLINED_SCENARIO_WITH_SUBSTITUTIONS_IN_TABLE)
    solved = solved_steps(scenario)
    for step, expected in zip(solved, expected_hashes_per_step):
        assert_equal(type(step), Step)
        assert_equal(step.hashes, expected)


def test_scenario_multilines_are_solved_against_outlines():
    """
    Outline substitution should apply to multiline strings within a scenario
    """

    expected_multiline = '<div>outline value</div>'

    scenario = parse_scenario(
        OUTLINED_SCENARIO_WITH_SUBSTITUTIONS_IN_MULTILINE)
    step = solved_steps(scenario)[0]

    assert_equal(type(step), Step)
    assert_equal(step.multiline, expected_multiline)


def test_solved_steps_also_have_scenario_as_attribute():
    """
    Steps solved in scenario outlines also have scenario as attribute
    """

    scenario = parse_scenario(OUTLINED_SCENARIO)
    for step in solved_steps(scenario):
        assert_equal(step.scenario, scenario)


def test_scenario_outlines_within_feature():
    """
    Solving scenario outlines within a feature
    """

    feature = Feature.from_string(OUTLINED_FEATURE)
    scenario = feature.scenarios[0]
    solved = solved_steps(scenario)

    assert_equal(len(solved), 12)
    expected_sentences = [
        'Given I have entered 20 into the calculator',
        'And I have entered 30 into the calculator',
        'When I press add',
        'Then the result should be 50 on the screen',
        'Given I have entered 2 into the calculator',
        'And I have entered 5 into the calculator',
        'When I press add',
        'Then the result should be 7 on the screen',
        'Given I have entered 0 into the calculator',
        'And I have entered 40 into the calculator',
        'When I press add',
        'Then the result should be 40 on the screen',
    ]

    for step, expected in zip(solved, expected_sentences):
        assert_equal(type(step), Step)
        assert_equal(step.sentence, expected)


def test_full_featured_feature():
    """
    Solving scenarios within a full-featured feature
    """

    feature = Feature.from_string(OUTLINED_FEATURE_WITH_MANY)
    scenario1, scenario2, scenario3, scenario4 = feature.scenarios

    assert_equal(scenario1.name, 'Do something')
    assert_equal(scenario2.name, 'Do something else')
    assert_equal(scenario3.name, 'Worked!')
    assert_equal(scenario4.name, 'Add two numbers wisely')

    solved = solved_steps(scenario1)

    assert_equal(len(solved), 2)
    expected_sentences = [
        'Given I have entered ok into the fail',
        'Given I have entered fail into the ok',
    ]
    for step, expected in zip(solved, expected_sentences):
        assert_equal(step.sentence, expected)

    expected_evaluated = (
        (
            {'button': 'add', 'input_1': '20',
             'input_2': '30', 'output': '50'},
            [
                'Given I have entered 20 into the calculator',
                'And I have entered 30 into the calculator',
                'When I press add',
                'Then the result should be 50 on the screen',
            ]
        ),
        (
            {'button': 'add', 'input_1': '2',
             'input_2': '5', 'output': '7'},
            [
                'Given I have entered 2 into the calculator',
                'And I have entered 5 into the calculator',
                'When I press add',
                'Then the result should be 7 on the screen',
            ]
        ),
        (
            {'button': 'add', 'input_1': '0',
             'input_2': '40', 'output': '40'},
            [
                'Given I have entered 0 into the calculator',
                'And I have entered 40 into the calculator',
                'When I press add',
                'Then the result should be 40 on the screen',
            ],
        ),
        (
            {'button': 'add', 'input_1': '5',
             'input_2': '7', 'output': '12'},
            [
                'Given I have entered 5 into the calculator',
                'And I have entered 7 into the calculator',
                'When I press add',
                'Then the result should be 12 on the screen',
            ],
        )
    )
    for ((got_examples, got_steps), (expected_examples, expected_steps)) \
            in zip(scenario4.evaluated, expected_evaluated):
        assert_equal(got_examples, expected_examples)
        assert_equal([x.sentence for x in got_steps],
                     expected_steps)


@skip("FIXME: Gherkin parser allows this")
def test_scenario_with_table_and_no_step_fails():
    "A step table imediately after the scenario line, without step line fails"

    with assert_raises(LettuceSyntaxError):
        parse_scenario(SCENARIO_FAILED)


def test_scenario_ignore_commented_lines_from_examples():
    "Comments on scenario example should be ignored"
    scenario = parse_scenario(OUTLINED_SCENARIO_WITH_COMMENTS_ON_EXAMPLES)

    assert_equal(
        scenario.outlines,
        (
            {'input_1': '20', 'input_2': '30',
             'button': 'add', 'output': '50'},
            {'input_1': '0', 'input_2': '40',
             'button': 'add', 'output': '40'},
        )
    )


def test_scenario_aggregate_all_examples_blocks():
    "All scenario's examples block should be translated to outlines"
    scenario = parse_scenario(
        OUTLINED_SCENARIO_WITH_MORE_THAN_ONE_EXAMPLES_BLOCK)

    assert_equal(
        scenario.outlines,
        (
            {'input_1': '20', 'input_2': '30',
             'button': 'add', 'output': '50'},
            {'input_1': '2', 'input_2': '5',
             'button': 'add', 'output': '7'},
            {'input_1': '0', 'input_2': '40',
             'button': 'add', 'output': '40'},
            {'input_1': '20', 'input_2': '33',
             'button': 'add', 'output': '53'},
            {'input_1': '12', 'input_2': '40',
             'button': 'add', 'output': '52'},
        )
    )


def test_commented_scenarios():
    "A scenario string that contains lines starting with '#' will be commented"
    scenario = parse_scenario(COMMENTED_SCENARIO)
    assert_equal(scenario.name,
                 'Adding some students to my university database')
    assert_equal(len(scenario.steps), 4)


def test_scenario_with_hash_within_double_quotes():
    ("Scenarios have hashes within double quotes and yet don't "
     "consider them as comments")

    scenario = parse_scenario(
        INLINE_COMMENTS_IGNORED_WITHIN_DOUBLE_QUOTES)

    step1, step2 = scenario.steps

    assert step1.sentence == u'Given I am logged in on twitter'
    assert step2.sentence == u'When I search for the hashtag "#hammer"'


def test_scenario_with_hash_within_single_quotes():
    ("Scenarios have hashes within single quotes and yet don't "
     "consider them as comments")

    scenario = parse_scenario(
        INLINE_COMMENTS_IGNORED_WITHIN_SINGLE_QUOTES)

    step1, step2 = scenario.steps

    assert step1.sentence == u'Given I am logged in on twitter'
    assert step2.sentence == u"When I search for the hashtag '#hammer'"



# -*- coding: utf-8 -*-
"""
Test French language parsing.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()

from nose.tools import assert_equal
from aloe.parser import Feature

SCENARIO = u"""
Scénario: Ajout de plusieurs cursus dans la base de mon université
    Soit Une liste de cursus disponibles dans mon université
        | Nom                       | Durée    |
        | Science de l'Informatique | 5 ans    |
        | Nutrition                 | 4 ans    |
    Quand je consolide la base dans 'cursus.txt'
    Alors je vois que la 1er ligne de 'cursus.txt' contient 'Science de l'Informatique:5
    Et je vois que la 2em ligne de 'cursus.txt' contient 'Nutrition:4'
"""  # noqa

OUTLINED_SCENARIO = u"""
Plan du Scénario: Ajouter 2 nombres
    Soit <input_1> entré dans la calculatrice
    Et <input_2> entré dans la calculatrice
    Quand je presse <bouton>
    Alors je doit avoir <output> à l'écran

    Exemples:
      | input_1 | input_2 | bouton | output |
      | 20      | 30      | add    | 50     |
      | 2       | 5       | add    | 7      |
      | 0       | 40      | add    | 40     |
"""

OUTLINED_SCENARIO2 = u"""
Plan du Scénario: Ajouter 2 nombres
    Soit <input_1> entré dans la calculatrice
    Et <input_2> entré dans la calculatrice
    Quand je presse <bouton>
    Alors je doit avoir <output> à l'écran

    Exemples:
      | input_1 | input_2 | bouton | output |
      | 20      | 30      | add    | 50     |
      | 2       | 5       | add    | 7      |
      | 0       | 40      | add    | 40     |
"""
OUTLINED_SCENARIO3 = u"""
Plan du Scénario: Ajouter 2 nombres
    Soit <input_1> entré dans la calculatrice
    Et <input_2> entré dans la calculatrice
    Quand je presse <bouton>
    Alors je doit avoir <output> à l'écran

    Exemples:
      | input_1 | input_2 | bouton | output |
      | 20      | 30      | add    | 50     |
      | 2       | 5       | add    | 7      |
      | 0       | 40      | add    | 40     |
"""

OUTLINED_FEATURE = u"""
Fonctionnalité: Faire plusieur choses en même temps
    De façon à automatiser les tests
    En tant que fainéant
    J'utilise les plans de scénario

    Plan du Scénario: Ajouter 2 nombres
        Soit <input_1> entré dans la calculatrice
        Et <input_2> entré dans la calculatrice
        Quand je presse <bouton>
        Alors je doit avoir <output> à l'écran

    Exemples:
        | input_1 | input_2 | bouton | output |
        | 20      | 30      | add    | 50     |
        | 2       | 5       | add    | 7      |
        | 0       | 40      | add    | 40     |
"""
OUTLINED_FEATURE2 = u"""
Fonctionnalité: Faire plusieur choses en même temps
    De façon à automatiser les tests
    En tant que fainéant
    J'utilise les plans de scénario

    Plan du Scénario: Ajouter 2 nombres
        Soit <input_1> entré dans la calculatrice
        Et <input_2> entré dans la calculatrice
        Quand je presse <bouton>
        Alors je doit avoir <output> à l'écran

    Exemples:
        | input_1 | input_2 | bouton | output |
        | 20      | 30      | add    | 50     |
        | 2       | 5       | add    | 7      |
        | 0       | 40      | add    | 40     |
"""


def parse_scenario(string, language=None):
    """Parse a scenario, prefixing it with a feature header."""
    feature = u"""
    Fonctionnalité: parse_scenario
    """

    feature += string
    feature = Feature.from_string(feature, language=language)

    return feature.scenarios[0]


def test_scenario_fr_from_string():
    """
    Language: FR -> Scenario.from_string
    """

    scenario = parse_scenario(SCENARIO, language='fr')

    assert_equal(
        scenario.name,
        u'Ajout de plusieurs cursus dans la base de mon université'
    )
    assert_equal(
        scenario.steps[0].hashes,
        (
            {'Nom': u"Science de l'Informatique", u'Durée': '5 ans'},
            {'Nom': u'Nutrition', u'Durée': '4 ans'},
        )
    )


def test_scenario_outline1_fr_from_string():
    """
    Language: FR -> Scenario.from_string, with scenario outline, first case
    """

    scenario = parse_scenario(OUTLINED_SCENARIO, language='fr')

    assert_equal(
        scenario.name,
        'Ajouter 2 nombres'
    )
    assert_equal(
        scenario.outlines,
        (
            {u'input_1': u'20', u'input_2': u'30',
             u'bouton': u'add', u'output': u'50'},
            {u'input_1': u'2', u'input_2': u'5',
             u'bouton': u'add', u'output': u'7'},
            {u'input_1': u'0', u'input_2': u'40',
             u'bouton': u'add', u'output': u'40'},
        )
    )


def test_scenario_outline2_fr_from_string():
    """
    Language: FR -> Scenario.from_string, with scenario outline, second case
    """

    scenario = parse_scenario(OUTLINED_SCENARIO2, language='fr')

    assert_equal(
        scenario.name,
        'Ajouter 2 nombres'
    )
    assert_equal(
        scenario.outlines,
        (
            {u'input_1': u'20', u'input_2': u'30',
             u'bouton': u'add', u'output': u'50'},
            {u'input_1': u'2', u'input_2': u'5',
             u'bouton': u'add', u'output': u'7'},
            {u'input_1': u'0', u'input_2': u'40',
             u'bouton': u'add', u'output': u'40'},
        )
    )


def test_scenario_outline3_fr_from_string():
    """
    Language: FR -> Scenario.from_string, with scenario outline, third case
    """

    scenario = parse_scenario(OUTLINED_SCENARIO2, language='fr')

    assert_equal(
        scenario.name,
        'Ajouter 2 nombres'
    )
    assert_equal(
        scenario.outlines,
        (
            {u'input_1': u'20', u'input_2': u'30',
             u'bouton': u'add', u'output': u'50'},
            {u'input_1': u'2', u'input_2': u'5',
             u'bouton': u'add', u'output': u'7'},
            {u'input_1': u'0', u'input_2': u'40',
             u'bouton': u'add', u'output': u'40'},
        )
    )


def test_feature_fr_from_string():
    """
    Language: FR -> Feature.from_string
    """

    feature = Feature.from_string(OUTLINED_FEATURE, language='fr')

    assert_equal(
        feature.name,
        u'Faire plusieur choses en même temps'
    )

    assert_equal(
        feature.description,
        u"De façon à automatiser les tests\n"
        u"En tant que fainéant\n"
        u"J'utilise les plans de scénario"
    )

    (scenario, ) = feature.scenarios

    assert_equal(
        scenario.name,
        'Ajouter 2 nombres'
    )

    assert_equal(
        scenario.outlines,
        (
            {u'input_1': u'20', u'input_2': u'30',
             u'bouton': u'add', u'output': u'50'},
            {u'input_1': u'2', u'input_2': u'5',
             u'bouton': u'add', u'output': u'7'},
            {u'input_1': u'0', u'input_2': u'40',
             u'bouton': u'add', u'output': u'40'},
        )
    )


def test_feature_fr_from_string2():
    """
    Language: FR -> Feature.from_string, alternate name
    """

    feature = Feature.from_string(OUTLINED_FEATURE2, language='fr')

    assert_equal(
        feature.name,
        u'Faire plusieur choses en même temps'
    )

    assert_equal(
        feature.description,
        u"De façon à automatiser les tests\n"
        u"En tant que fainéant\n"
        u"J'utilise les plans de scénario"
    )

    (scenario, ) = feature.scenarios

    assert_equal(
        scenario.name,
        'Ajouter 2 nombres'
    )

    assert_equal(
        scenario.outlines,
        (
            {u'input_1': u'20', u'input_2': u'30',
             u'bouton': u'add', u'output': u'50'},
            {u'input_1': u'2', u'input_2': u'5',
             u'bouton': u'add', u'output': u'7'},
            {u'input_1': u'0', u'input_2': u'40',
             u'bouton': u'add', u'output': u'40'},
        )
    )



"""
Miscellaneous utilities used internally by Aloe.
"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin, unused-wildcard-import, wildcard-import
from builtins import *
# pylint:enable=redefined-builtin, unused-wildcard-import, wildcard-import
from future import standard_library
standard_library.install_aliases()

import io
import re
import sys

try:
    from functools import lru_cache  # pylint:disable=no-name-in-module
except ImportError:
    from repoze.lru import lru_cache
from contextlib import contextmanager


@contextmanager
def dummy_cm():
    """A dummy context manager to compare other decorated functions to."""
    pass


PY3 = sys.version_info >= (3, 0)


def identifier(value):
    """
    Make a valid identifier from a string on either Python 2 or Python 3.

    Python 2 limits identifiers to ASCII characters, so unicode must be
    encoded.
    """

    if PY3:
        return value
    else:
        return value.encode('unicode_escape')


if PY3:
    TestWrapperIO = io.StringIO  # pylint:disable=invalid-name
else:
    class TestWrapperIO(io.StringIO):
        """A wrapper for capturing Nose output in tests."""

        def write(self, str_):
            """
            Write a string to the stream. In case of Python 2, accept both str
            and unicode.
            """

            try:
                super().write(str_)
            except TypeError:
                super().write(str_.decode('utf-8'))

        # The following methods are added only so Pylint can infer them

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def getvalue(self):
            """Get the captured string."""
            return super().getvalue()


def unwrap_function(func):
    """
    Given a function to which a decorator was applied, return the original
    function.
    """

    if hasattr(func, '__wrapped__'):
        return func.__wrapped__

    # In Python 2, functools.wraps doesn't set __wrapped__, however, it is
    # possible to check whether the function has been wrapped by a known
    # method, in this case contextlib.contextmanager
    elif func.__code__.co_filename == dummy_cm.__code__.co_filename and \
            func.__code__.co_firstlineno == dummy_cm.__code__.co_firstlineno:
        # Get the original function from the closure
        return func.__closure__[0].cell_contents

    else:
        return func


RE_CAMEL_CASE = re.compile(r'(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))')


def camel_case_to_spaces(value):
    """
    Splits CamelCase and converts to lower case. Also strips leading and
    trailing whitespace.
    """
    return RE_CAMEL_CASE.sub(r' \1', value).strip().lower()


class memoizedproperty(object):  # pylint:disable=invalid-name
    """
    A property that is only computed on the first access.
    """

    def __init__(self, func):
        self.func = func
        self.name = func.__name__

    def __get__(self, instance, owner):
        """Compute the value and cache it in the class dict."""

        if instance is None:
            return self

        result = self.func(instance)
        instance.__dict__[self.name] = result
        return result


class memoizedtype(type):  # pylint:disable=invalid-name
    """
    A type that caches the created instances.
    """

    @lru_cache(20)
    def __call__(cls, *args, **kwargs):
        # On Python 2, newsuper can't deal with metaclasses
        return super(memoizedtype, cls).__call__(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
A Gherkin parser written using pyparsing.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin,wildcard-import,unused-wildcard-import
from builtins import *
# pylint:enable=redefined-builtin,wildcard-import,unused-wildcard-import
from future import standard_library
standard_library.install_aliases()

import os
from collections import OrderedDict
from copy import copy
from io import StringIO

from gherkin3.dialect import Dialect
from gherkin3.parser import Parser
from gherkin3.token_matcher import TokenMatcher
from gherkin3.token_scanner import TokenScanner as BaseTokenScanner

from aloe import strings
from aloe.exceptions import LettuceSyntaxError
from aloe.utils import memoizedproperty

# Pylint can't figure out methods vs. properties and which classes are
# abstract
# pylint:disable=abstract-method


class TokenScanner(BaseTokenScanner):
    """Gherkin 3 token scanner that explicitly takes a string or a filename."""

    # pylint:disable=super-init-not-called
    def __init__(self, string=None, filename=None):
        if string:
            if filename:
                raise ValueError(
                    "Cannot provide string and filename together.")
            self.io = StringIO(string)  # pylint:disable=invalid-name
        elif filename:
            self.io = open(filename, 'r')
        else:
            raise ValueError("Must provide either string or filename.")

        self.line_number = 0
    # pylint:enable=super-init-not-called


class LanguageTokenMatcher(TokenMatcher):
    """Gherkin 3 token matcher that always uses the given language."""

    def __init__(self, dialect_name='en'):
        self.actual_dialect_name = dialect_name
        super().__init__(dialect_name=dialect_name)

    def _change_dialect(self, dialect_name, location=None):
        """Force the dialect name given in the constructor."""
        super()._change_dialect(self.actual_dialect_name, location=location)


def cell_values(row):
    """Extract cell values from a table header or row."""

    return tuple(cell['value'] for cell in row['cells'])


class Node(object):
    """
    A base parse node.
    """

    def __init__(self, parsed, filename=None):
        """Construct the node from Gherkin parse results."""
        self.line = parsed['location']['line']
        self.col = parsed['location']['column']
        self.filename = filename

    @property
    def feature(self):
        """The feature this node ultimately belongs to."""

        raise NotImplementedError

    @property
    def location(self):
        """Location as 'filename:line'"""

        return '{filename}:{line}'.format(
            filename=os.path.relpath(self.filename),
            line=self.line,
        )

    @property
    def text(self):
        """The text for this node."""

        raise NotImplementedError

    indent = 0  # The indent to use when printing the node

    def represented(self):
        """A representation of the node."""

        result = ' ' * self.indent + self.text.strip()

        return result


class Step(Node):
    """
    A single statement within a test.

    A :class:`Scenario` or :class:`Background` is composed of multiple
    :class:`Step`.
    """

    table = None
    """
    A Gherkin table as an iterable of rows, themselves iterables of cells.

    e.g.:

    .. code-block:: gherkin

        Then I have fruit:
            | apples | oranges |
            | 0      | 2       |

    Becomes:

    .. code-block:: python

        (('apples', 'oranges'), ('0', '2'))
    """

    multiline = None
    """
    A Gherkin multiline string with the appropriate indenting removed.

    .. code-block:: gherkin

        Then I have poem:
            \"\"\"
            Glittering-Minded deathless Aphrodite,
            I beg you, Zeus’s daughter, weaver of snares,
            Don’t shatter my heart with fierce
            Pain, goddess,
            \"\"\"
    """

    def __init__(self, parsed, background=None, scenario=None, **kwargs):
        super().__init__(parsed, **kwargs)

        if background:
            self.background = background
        elif scenario:
            self.scenario = scenario
        else:
            raise ValueError(
                "Step must belong to either a scenario or a background.")

        self.sentence = parsed['keyword'] + parsed['text']
        """The sentence parsed for this step."""

        try:
            argument_type = parsed['argument']['type']
        except KeyError:
            argument_type = None

        if argument_type == 'DataTable':
            self.table = tuple(
                cell_values(row)
                for row in parsed['argument']['rows']
            )

        elif argument_type == 'DocString':
            self.multiline = parsed['argument']['content']

    @property
    def text(self):
        return self.sentence

    def __str__(self):
        return '<Step: "%s">' % self.sentence

    def __repr__(self):
        return str(self)

    def parse_steps_from_string(self, string, **kwargs):
        """
        Parse a number of steps, returns an iterable of :class:`Step`.

        This is used by :func:`step.behave_as`.
        """

        try:
            container = self.scenario
            is_scenario = True
        except AttributeError:
            container = self.background
            is_scenario = False

        # Gherkin can't parse anything other than complete features
        feature_string = """
        # language: {feature.language}
        {feature.keyword}: feature

        {container.keyword}: scenario
        {string}
        """.format(
            feature=self.feature,
            container=container,
            string=string,
        )

        feature = self.feature.parse(string=feature_string,
                                     filename=self.filename)

        if is_scenario:
            return feature.scenarios[0].steps
        else:
            return feature.background.steps

    @property
    def feature(self):
        """
        The :class:`Feature` this step is a part of.
        """

        try:
            return self.scenario.feature
        except AttributeError:
            return self.background.feature

    @memoizedproperty
    def keys(self):
        """
        Return the first row of a table if this statement contains one.
        """
        if self.table:
            return tuple(self.table[0])
        else:
            return ()

    @memoizedproperty
    def hashes(self):
        """
        Return the table attached to the step as an iterable of hashes, where
        the first row - the column headings - supplies keys for all the others.

        e.g.:

        .. code-block:: gherkin

            Then I have fruit:
                | apples | oranges |
                | 0      | 2       |

        Becomes:

        .. code-block:: python

            ({
                'apples': '0',
                'oranges': '2',
            },)
        """

        if not self.table:
            return ()

        keys = self.keys

        return tuple(
            dict(zip(keys, row))
            for row in self.table[1:]
        )

    @memoizedproperty
    def max_length(self):
        """
        The max length of the feature, description and child blocks
        """

        return max(
            0,
            strings.get_terminal_width(self.represented(table=False,
                                                        multiline=False)),
            *[strings.get_terminal_width(line)
              for line in self.represent_table().splitlines()]
        )

    indent = 4

    # pylint:disable=arguments-differ
    def represented(self, table=True, multiline=True, color=str):
        """
        Render the line.
        """

        lines = [color(super().represented())]

        if table and self.table:
            lines.append(self.represent_table(cell_wrap=color))

        if multiline and self.multiline:
            lines.append(self.represent_multiline(string_wrap=color))

        return '\n'.join(lines)
    # pylint:enable=arguments-differ

    def represent_table(self, **kwargs):
        """
        Render the table.

        :param cell_wrap: color to use inside the table cells
        """

        return strings.represent_table(
            self.table, indent=self.indent + 2, **kwargs)

    def represent_multiline(self, string_wrap=str):
        """
        Render the multiline.

        :param string_wrap: color to use inside the string
        """

        indent = self.indent + 2

        lines = [' ' * indent + '"""']
        lines += [' ' * indent + string_wrap(line)
                  for line in self.multiline.splitlines()]
        lines += [' ' * indent + '"""']

        return '\n'.join(lines)

    def resolve_substitutions(self, outline):
        """
        Creates a copy of the step with any <variables> resolved.
        """

        replaced = copy(self)

        def replace_vars(string):
            """Replace all the variables in a string."""
            for key, value in outline.items():
                key = '<{key}>'.format(key=key)
                string = string.replace(key, value)
            return string

        replaced.sentence = replace_vars(self.sentence)

        if self.multiline:
            replaced.multiline = replace_vars(self.multiline)

        if self.table:
            replaced.table = tuple(
                tuple(
                    replace_vars(cell)
                    for cell in row
                )
                for row in self.table
            )

        return replaced

    def step_keyword(self, kind):
        """
        An appropriate keyword for a particular kind of step
        (Given, When, Then) for the language the current step is written in.
        """

        dialect = self.feature.dialect
        keywords = {
            'given': dialect.given_keywords,
            'when': dialect.when_keywords,
            'then': dialect.then_keywords,
        }[kind]()

        # Gherkin allows '*' as a keyword; skip it to be sure the keyword is
        # specifically for the given kind
        return next(
            keyword for keyword in keywords
            if not keyword.startswith('*')
        )


class StepContainer(Node):
    """A node containing steps, e.g. Feature:, Scenario:"""

    step_class = Step

    container_name = 'container'  # override in subclasses

    @property
    def feature(self):
        return self._feature

    def __init__(self, parsed, feature=None, filename=None, **kwargs):
        super().__init__(parsed, filename=filename, **kwargs)

        self._feature = feature

        # Put a reference to the parent node into all the steps
        parent_ref = {self.container_name: self}

        self.steps = tuple(
            self.step_class(step, filename=filename, **parent_ref)
            for step in parsed['steps']
        )

    indent = 2


class HeaderNode(Node):
    """
    Nodes with a header consisting of a keyword, name and a number of tags.
    """

    def __init__(self, parsed, **kwargs):
        super().__init__(parsed, **kwargs)

        self._tags = tuple(
            tag['name'][1:] for tag in parsed['tags']
        )
        self.keyword = parsed['keyword']
        self.name = parsed['name'].strip()

        if self.name == '':
            raise LettuceSyntaxError(
                None,
                "{line}:{col} {klass} must have a name".format(
                    line=self.line,
                    col=self.col,
                    klass=self.__class__.__name__))

    def __str__(self):
        return '<{klass}: "{name}">'.format(
            # FIXME: use self.language
            klass=self.__class__.__name__,
            name=self.name)

    def __repr__(self):
        return str(self)

    @property
    def tags(self):
        """
        Tags for a feature.

        Tags are applied to a feature using the appropriate Gherkin syntax:

        .. code-block:: gherkin

            @tag1 @tag2
            Feature: Eat leaves
        """
        return self._tags

    @property
    def text(self):
        """The text for this block."""

        return '{keyword}: {name}'.format(keyword=self.keyword,
                                          name=self.name)

    def represent_tags(self):
        """
        Render the tags of a tagged block.
        """

        return ' ' * self.indent + '  '.join('@%s' % tag for tag in self.tags)


class Background(StepContainer):
    """The background of all :class:`Scenario` in a :class:`Feature`."""
    container_name = 'background'

    text = 'Background:'


class Outline(OrderedDict, Node):
    """An outline within a :class:`Scenario`."""

    def __init__(self, keys, table_row, filename=None):
        """Construct the outline."""

        # Extract values
        OrderedDict.__init__(self, zip(keys, cell_values(table_row)))

        # Store the file and line information
        Node.__init__(self, table_row, filename=filename)


class Scenario(HeaderNode, StepContainer):
    """A scenario within a :class:`Feature`."""

    container_name = 'scenario'

    def __init__(self, parsed, **kwargs):
        super().__init__(parsed, **kwargs)

        # Build a list of outline hashes
        # A single scenario can have multiple example blocks, the returned
        # token is a list of table tokens
        self.outlines = ()

        for example_table in parsed.get('examples', ()):
            # the first row of the table is the column headings
            keys = cell_values(example_table['tableHeader'])

            self.outlines += tuple(
                Outline(keys, row)
                for row in example_table['tableBody']
            )

    indent = 2

    def represent_outlines(self):
        """
        Render the outlines table.
        """

        return strings.represent_table(
            self.outlines_table, indent=self.indent + 2)

    @memoizedproperty
    def max_length(self):
        """
        The max horizontal length of the feature, description and child blocks.
        """

        return max(
            0,
            strings.get_terminal_width(self.represented()),
            *([step.max_length for step in self.steps] +
              [strings.get_terminal_width(line)
               for line in self.represent_outlines().splitlines()])
        )

    @memoizedproperty
    def outlines_table(self):
        """
        Return the outlines as a table.
        """

        # get the list of column headings
        headings = OrderedDict()

        for outline in self.outlines:
            headings.update(outline)

        headings = list(headings.keys())
        table = [headings]

        # append the data to the table
        for outline in self.outlines:
            table.append([outline.get(cell, '') for cell in headings])

        return table

    @property
    def tags(self):
        """Tags for the :attr:`feature` and the scenario."""
        return self._tags + self.feature.tags

    @property
    def evaluated(self):
        """
        Yield the outline and steps.
        """

        for outline in self.outlines:
            steps = [step.resolve_substitutions(outline)
                     for step in self.steps]

            # set a backref to the scenario
            for step in steps:
                step.scenario = self

            yield (outline, steps)


class Description(Node):
    """
    The description block of a feature.
    """

    def __init__(self, parsed, **kwargs):
        super().__init__(parsed, **kwargs)

        description = parsed.get('description', '')
        self.lines = tuple(line.strip() for line in description.split('\n'))

    def __str__(self):
        return '\n'.join(self.lines)

    def __repr__(self):
        return str(self)

    indent = 2

    def represented(self):
        return '\n'.join(
            self.represent_line(n)
            for n, _ in enumerate(self.lines)
        )

    def represent_line(self, idx):
        """
        Render the nth line in the description.
        """

        line = self.lines[idx]
        result = ' ' * self.indent + line

        return result

    @memoizedproperty
    def description_at(self):
        """
        Return a tuple of lines in the string containing the description.
        """

        offset = self.line

        return tuple(offset + lineno for lineno, _
                     in enumerate(self.lines))

    @memoizedproperty
    def max_length(self):
        """
        The maximum length of all description lines.
        """
        try:
            return max(
                strings.get_terminal_width(self.represent_line(n))
                for n, _ in enumerate(self.lines)
            )
        except ValueError:
            return 0


class Feature(HeaderNode):
    """
    A complete Gherkin feature.

    Features can either be constructed :func:`from_file` or
    :func:`from_string`.
    """

    background_class = Background
    scenario_class = Scenario

    background = None

    def __init__(self, parsed, filename=None, **kwargs):
        super().__init__(parsed, filename=filename, **kwargs)

        self.language = parsed['language']

        self.description_node = Description(parsed, filename=filename)

        if 'background' in parsed:
            self.background = self.background_class(parsed['background'],
                                                    filename=filename,
                                                    feature=self)

        self.scenarios = tuple(
            self.scenario_class(scenario, filename=filename, feature=self)
            for scenario in parsed['scenarioDefinitions']
        )

    @classmethod
    def parse(cls, string=None, filename=None, language=None):
        """
        Parse either a string or a file.
        """

        parser = Parser()
        if language:
            if language == 'pt-br':
                language = 'pt'
            token_matcher = LanguageTokenMatcher(language)
        else:
            token_matcher = TokenMatcher()

        if string:
            token_scanner = TokenScanner(string=string)
        else:
            token_scanner = TokenScanner(filename=filename)

        return cls(
            parser.parse(token_scanner, token_matcher=token_matcher),
            filename=filename,
        )

    @classmethod
    def from_string(cls, string, language=None):
        """
        Parse a string into a :class:`Feature`.
        """

        return cls.parse(string=string, language=language)

    @classmethod
    def from_file(cls, filename, language=None):
        """
        Parse a file or filename into a :class:`Feature`.
        """

        return cls.parse(filename=filename, language=language)

    @property
    def description(self):
        """
        The description of the feature (the text that comes directly under
        the feature).
        """
        return str(self.description_node)

    @property
    def dialect(self):
        """
        The Gherkin dialect for the feature.
        """

        return Dialect.for_name(self.language)

    @property
    def feature(self):
        """
        Convenience property for generic functions.
        """

        return self

    @memoizedproperty
    def max_length(self):
        """
        The max horizontal length of the feature, description and child blocks.

        This is used for aligning rendered output.
        """

        return max(
            0,
            strings.get_terminal_width(
                self.represented(description=False)),
            self.description_node.max_length,
            *[scenario.max_length for scenario in self.scenarios]
        )

    # pylint:disable=arguments-differ
    def represented(self, description=True):
        result = super().represented()

        if description and self.description != '':
            result += '\n'
            result += self.description_node.represented()

        return result



# -*- coding: utf-8 -*-
"""
Test step and callback registry.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()

import re
import unittest

from nose.tools import (
    assert_equal,
    assert_in,
    assert_not_in,
    assert_raises,
)

from aloe.registry import (
    CallbackDecorator,
    CallbackDict,
    PriorityClass,
    StepDict,
)
from aloe.exceptions import StepLoadingError

from tests.utils import appender, before_after


def test_StepDict_raise_StepLoadingError_if_first_argument_is_not_a_regex():
    """
    aloe.STEP_REGISTRY.load(step, func) should raise an error if step is
    not a regex
    """
    steps = StepDict()
    with assert_raises(StepLoadingError):
        steps.load("an invalid regex;)", lambda: "")


def test_StepDict_can_load_a_step_composed_of_a_regex_and_a_function():
    """
    aloe.STEP_REGISTRY.load(step, func) append item(step, func) to
    STEP_REGISTRY
    """
    steps = StepDict()

    def func():  # pylint:disable=missing-docstring
        return ""

    step = "a step to test"
    steps.load(step, func)

    step = re.compile(step, re.I | re.U)
    assert_in(step, steps)
    assert_equal(steps[step], func)


def test_StepDict_load_a_step_return_the_given_function():
    """
    aloe.STEP_REGISTRY.load(step, func) returns func
    """
    steps = StepDict()

    def func():  # pylint:disable=missing-docstring
        return ""

    assert_equal(steps.load("another step", func), func)


def test_StepDict_can_extract_a_step_sentence_from_function_name():
    """
    aloe.STEP_REGISTRY.extract_sentence(func) parse func name and return
    a sentence
    """
    steps = StepDict()

    def a_step_sentence():  # pylint:disable=missing-docstring
        pass
    assert_equal("A step sentence", steps.extract_sentence(a_step_sentence))


def test_StepDict_can_extract_a_step_sentence_from_function_doc():
    """
    aloe.STEP_REGISTRY.extract_sentence(func) parse func doc and return
    a sentence
    """
    steps = StepDict()

    def a_step_func():
        """A step sentence"""
        pass
    assert_equal("A step sentence", steps.extract_sentence(a_step_func))


def test_StepDict_can_load_a_step_from_a_function():
    """
    aloe.STEP_REGISTRY.load_func(func) append item(step, func) to
    STEP_REGISTRY
    """
    steps = StepDict()

    def a_step_to_test():  # pylint:disable=missing-docstring
        pass

    steps.load_func(a_step_to_test)

    expected_sentence = re.compile("A step to test", re.I | re.U)
    assert_in(expected_sentence, steps)
    assert_equal(steps[expected_sentence], a_step_to_test)


def test_StepDict_can_load_steps_from_an_object():
    """
    aloe.STEP_REGISTRY.load_steps(obj) append all obj methods to
    STEP_REGISTRY
    """
    steps = StepDict()

    class LotsOfSteps(object):
        """A class defining some steps."""

        def step_1(self):  # pylint:disable=missing-docstring
            pass

        def step_2(self):
            """Doing something"""
            pass

    step_list = LotsOfSteps()
    steps.load_steps(step_list)

    expected_sentence1 = re.compile("Step 1", re.I | re.U)
    expected_sentence2 = re.compile("Doing something", re.I | re.U)
    assert_in(expected_sentence1, steps)
    assert_in(expected_sentence2, steps)
    assert_equal(steps[expected_sentence1], step_list.step_1)
    assert_equal(steps[expected_sentence2], step_list.step_2)


def test_StepDict_can_exclude_methods_when_load_steps():
    """
    aloe.STEP_REGISTRY.load_steps(obj) don't load exluded attr in
    STEP_REGISTRY
    """
    steps = StepDict()

    class LotsOfSteps(object):
        """A class defining some steps."""
        exclude = ["step_1"]

        def step_1(self):  # pylint:disable=missing-docstring
            pass

        def step_2(self):
            """Doing something"""
            pass

    step_list = LotsOfSteps()
    steps.load_steps(step_list)

    expected_sentence1 = re.compile("Step 1", re.I | re.U)
    expected_sentence2 = re.compile("Doing something", re.I | re.U)
    assert_not_in(expected_sentence1, steps)
    assert_in(expected_sentence2, steps)


def test_StepDict_can_exclude_callable_object_when_load_steps():
    """
    aloe.STEP_REGISTRY.load_steps(obj) don't load callable objets in
    STEP_REGISTRY
    """
    steps = StepDict()

    class NoStep(object):
        """A class defining something that's not a step."""
        class NotAStep(object):
            """A callable which isn't a step."""
            def __call__(self):
                pass

    no_step = NoStep()
    steps.load_steps(no_step)

    assert len(steps) == 0


def test_unload_reload():
    """
    Test unloading and then reloading the step.
    """

    def step():  # pylint:disable=missing-docstring
        pass

    class StepDefinition(object):
        """A step definition object to match."""
        sentence = 'My step 1'

    steps = StepDict()

    # Load
    steps.step(r'My step (\d)')(step)

    assert len(steps) == 1
    assert steps.match_step(StepDefinition) == (step, ('1',), {})

    # Members added to step by registering it
    # pylint:disable=no-member

    # Unload
    step.unregister()

    assert len(steps) == 0

    # Should be a no-op
    step.unregister()

    assert len(steps) == 0

    # Reload
    steps.step(step.sentence)(step)

    assert len(steps) == 1
    assert steps.match_step(StepDefinition) == (step, ('1',), {})


class CallbackDictTest(unittest.TestCase):
    """
    Test callback dictionary.
    """

    def setUp(self):
        self.callbacks = CallbackDict()

        self.before = CallbackDecorator(self.callbacks, 'before')
        self.around = CallbackDecorator(self.callbacks, 'around')
        self.after = CallbackDecorator(self.callbacks, 'after')

    def test_wrap(self):
        """
        Test wrapping functions.
        """

        sequence = []

        self.before.all(appender(sequence, 'before'))

        self.around.all(before_after(
            appender(sequence, 'around_before'),
            appender(sequence, 'around_after')
        ))

        self.after.all(appender(sequence, 'after'))

        wrapped = appender(sequence, 'wrapped')

        wrap = self.callbacks.wrap('all', wrapped, 'hook_arg1', 'hook_arg2')

        wrap('wrap_arg1', 'wrap_arg2')

        self.assertEqual(sequence, [
            ('before', 'hook_arg1', 'hook_arg2'),
            ('around_before', 'hook_arg1', 'hook_arg2'),
            ('wrapped', 'wrap_arg1', 'wrap_arg2'),
            ('around_after', 'hook_arg1', 'hook_arg2'),
            ('after', 'hook_arg1', 'hook_arg2'),
        ])

    def test_before_after(self):
        """
        Test before_after.
        """

        sequence = []

        self.before.all(appender(sequence, 'before'))

        self.around.all(before_after(
            appender(sequence, 'around_before'),
            appender(sequence, 'around_after')
        ))

        self.after.all(appender(sequence, 'after'))

        before, after = self.callbacks.before_after('all')

        before('before_arg1', 'before_arg2')
        after('after_arg1', 'after_arg2')

        self.assertEqual(sequence, [
            ('before', 'before_arg1', 'before_arg2'),
            ('around_before', 'before_arg1', 'before_arg2'),
            ('around_after', 'before_arg1', 'before_arg2'),
            ('after', 'after_arg1', 'after_arg2'),
        ])

    @staticmethod
    def before_after_hook(sequence, when):
        """A before/after hook appending to a sequence."""
        return lambda name: appender(sequence, when + name)

    @classmethod
    def around_hook(cls, sequence):
        """An around hook appending to a sequence."""
        return lambda name: before_after(
            cls.before_after_hook(sequence, 'around_before')(name),
            cls.before_after_hook(sequence, 'around_after')(name)
        )

    def test_priority(self):
        """
        Test callback priority.
        """

        self.maxDiff = None

        sequence = []

        for when in ('before', 'after', 'around'):
            add_callback = getattr(self, when).all
            if when == 'around':
                hook = self.around_hook(sequence)
            else:
                hook = self.before_after_hook(sequence, when)

            # Default priority is 0
            add_callback(hook('B1'))
            add_callback(hook('B2'))

            # Explicit lower (=earlier) priority
            add_callback(hook('A1'), priority=-10)
            add_callback(hook('A2'), priority=-10)

            # Explicit higher (=later) priority
            add_callback(hook('C1'), priority=10)
            add_callback(hook('C2'), priority=10)

            # Add a callback with a different priority class
            CallbackDecorator(self.callbacks, when,
                              priority_class=-1).all(hook('Z1'))
            CallbackDecorator(self.callbacks, when,
                              priority_class=1).all(hook('D1'))

        wrap = self.callbacks.wrap('all', appender(sequence, 'wrapped'))

        wrap()

        self.assertEqual([item for (item,) in sequence], [
            'beforeZ1',
            'beforeA1',
            'beforeA2',
            'beforeB1',
            'beforeB2',
            'beforeC1',
            'beforeC2',
            'beforeD1',

            'around_beforeZ1',
            'around_beforeA1',
            'around_beforeA2',
            'around_beforeB1',
            'around_beforeB2',
            'around_beforeC1',
            'around_beforeC2',
            'around_beforeD1',

            'wrapped',

            'around_afterD1',
            'around_afterC2',
            'around_afterC1',
            'around_afterB2',
            'around_afterB1',
            'around_afterA2',
            'around_afterA1',
            'around_afterZ1',

            'afterD1',
            'afterC2',
            'afterC1',
            'afterB2',
            'afterB1',
            'afterA2',
            'afterA1',
            'afterZ1',
        ])

    def test_clear(self):
        """
        Test clearing the registry.
        """

        def prepare_hooks():
            """Set up various hooks to test clearing only some of them."""
            callbacks = CallbackDict()
            sequence = []

            for when in ('before', 'after', 'around'):
                add_callback = CallbackDecorator(callbacks, when).all
                if when == 'around':
                    hook = self.around_hook(sequence)
                else:
                    hook = self.before_after_hook(sequence, when)

                # Default priority class
                add_callback(hook('Default'))

                # Default priority class, specifying a name
                add_callback(hook('Named'), name='named')

                # Different priority classes
                CallbackDecorator(callbacks, when,
                                  priority_class=-1).all(hook('Minus'))
                CallbackDecorator(callbacks, when,
                                  priority_class=1).all(hook('Plus'))

                # Different priority class, specifying a name
                CallbackDecorator(callbacks, when,
                                  priority_class=1).all(hook('PlusNamed'),
                                                        name='named')

            return callbacks, sequence

        # Verify ordering without clearing anything
        callbacks, sequence = prepare_hooks()
        callbacks.wrap('all', appender(sequence, 'wrapped'))()

        self.assertEqual([item for (item,) in sequence], [
            'beforeMinus',
            'beforeDefault',
            'beforeNamed',
            'beforePlus',
            'beforePlusNamed',

            'around_beforeMinus',
            'around_beforeDefault',
            'around_beforeNamed',
            'around_beforePlus',
            'around_beforePlusNamed',

            'wrapped',

            'around_afterPlusNamed',
            'around_afterPlus',
            'around_afterNamed',
            'around_afterDefault',
            'around_afterMinus',

            'afterPlusNamed',
            'afterPlus',
            'afterNamed',
            'afterDefault',
            'afterMinus',
        ])

        # Only clear a particular name from the default priority class
        callbacks, sequence = prepare_hooks()
        callbacks.clear(priority_class=PriorityClass.USER,
                        name='named')
        callbacks.wrap('all', appender(sequence, 'wrapped'))()

        self.assertEqual([item for (item,) in sequence], [
            'beforeMinus',
            'beforeDefault',
            'beforePlus',
            'beforePlusNamed',

            'around_beforeMinus',
            'around_beforeDefault',
            'around_beforePlus',
            'around_beforePlusNamed',

            'wrapped',

            'around_afterPlusNamed',
            'around_afterPlus',
            'around_afterDefault',
            'around_afterMinus',

            'afterPlusNamed',
            'afterPlus',
            'afterDefault',
            'afterMinus',
        ])

        # Only clear the default priority class
        callbacks, sequence = prepare_hooks()
        callbacks.clear(priority_class=PriorityClass.USER)
        callbacks.wrap('all', appender(sequence, 'wrapped'))()

        self.assertEqual([item for (item,) in sequence], [
            'beforeMinus',
            'beforePlus',
            'beforePlusNamed',

            'around_beforeMinus',
            'around_beforePlus',
            'around_beforePlusNamed',

            'wrapped',

            'around_afterPlusNamed',
            'around_afterPlus',
            'around_afterMinus',

            'afterPlusNamed',
            'afterPlus',
            'afterMinus',
        ])

        # Clear all callbacks
        callbacks, sequence = prepare_hooks()
        callbacks.clear()
        callbacks.wrap('all', appender(sequence, 'wrapped'))()

        self.assertEqual([item for (item,) in sequence], [
            'wrapped',
        ])



"""
Test utility functions.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from future.utils import with_metaclass

import sys
import unittest

from contextlib import contextmanager
from functools import wraps

from aloe.utils import (
    unwrap_function,
    memoizedproperty,
    memoizedtype,
)


def my_function():
    """A function to test unwrapping on."""
    pass  # pragma: no cover


class UnwrapTest(unittest.TestCase):
    """
    Test unwrapping functions.
    """

    def test_plain_function(self):
        """Test unwrapping the raw function."""

        self.assertEqual(unwrap_function(my_function), my_function)

    def test_wraps(self):
        """Test unwrapping a function from a decorator using wraps."""

        if sys.version_info < (3, 0):
            # In Python 2, functools doesn't set enough information on the
            # wrapped function
            raise unittest.SkipTest(
                "Cannot unwrap @wraps in general case on Python 2.")

        @wraps(my_function)
        def decorated():
            """A decorated function."""
            return my_function()

        self.assertEqual(unwrap_function(decorated), my_function)

    def test_contextmanager(self):
        """Test unwrapping a context manager."""

        self.assertEqual(
            unwrap_function(contextmanager(my_function)),
            my_function
        )


class MemoizedTest(unittest.TestCase):
    """Test memoization functions."""

    def test_memoizedproperty(self):
        """Test memoizedproperty."""

        class Memoized(object):
            """A class to test memoizedproperty."""

            def __init__(self, value):
                self.value = value

            @memoizedproperty
            def prop(self):
                """
                A property that intentionally modifies the instance state on
                every access.
                """

                self.value += 1
                return self.value

        first = Memoized(5)
        second = Memoized(10)

        # Test that multiple accesses don't result in multiple calls
        self.assertEqual(first.prop, 6)
        self.assertEqual(first.prop, 6)

        # Test that different objects aren't sharing the value
        self.assertEqual(second.prop, 11)
        self.assertEqual(second.prop, 11)

    def test_memoizedtype(self):
        """Test memoizedtype."""

        class Memoized(with_metaclass(memoizedtype, object)):
            """A class to test memoizedtype."""

            counter = 0

            def __init__(self, value):
                """
                A constructor that intentionally modifies the class state.
                """

                self.value = value
                type(self).counter += 1

        first = Memoized(5)
        second = Memoized(10)

        # Check the objects are created properly
        self.assertEqual(first.value, 5)
        self.assertEqual(second.value, 10)

        # Try to create the same object again
        another_first = Memoized(5)

        # It should be the same as before
        self.assertIs(first, another_first)

        # Only two objects should have been created
        self.assertEqual(Memoized.counter, 2)



"""
Steps for features used in unit tests.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from aloe import step


@step(r'I do nothing')
def trivial_step(self):  # pylint:disable=unused-argument
    """Trivial passing step."""
    pass


@step(r'I fail')
def failing_step(self):  # pylint:disable=unused-argument
    """Trivial failing step."""
    raise AssertionError("This step is meant to fail.")



"""
Part of steps for the tested application.
"""

# pylint:disable=unused-argument,unused-import

from aloe import step

from .one import step_one


@step(r'I use step two')
def step_two(self):
    """Dummy step."""
    pass



"""
Test Factory Boy steps.
"""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
# pylint:disable=redefined-builtin, unused-wildcard-import, wildcard-import
from builtins import *
# pylint:enable=redefined-builtin, unused-wildcard-import, wildcard-import
from future import standard_library
standard_library.install_aliases()

from aloe.testing import (
    FeatureTest,
    in_directory,
)


@in_directory('tests/factoryboy_app')
class FactoryBoyTest(FeatureTest):
    """
    Test Factory Boy steps.
    """

    def test_factoryboy(self):
        """
        Test Factory Boy steps.
        """

        self.assert_feature_success('features/good.feature')
        self.assert_feature_fail('features/bad.feature')



"""
Cucumber-esque outputter for Nose.
"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin, unused-wildcard-import, wildcard-import
from builtins import *
# pylint:enable=redefined-builtin, unused-wildcard-import, wildcard-import
from future import standard_library
standard_library.install_aliases()

from contextlib import contextmanager
from functools import wraps

import blessings
from aloe.registry import (
    CallbackDecorator,
    CALLBACK_REGISTRY,
    PriorityClass,
)
from aloe.strings import ljust, represent_table
from aloe.tools import hook_not_reentrant
from nose.result import TextTestResult

# A decorator to add callbacks which wrap the steps looser than all the other
# callbacks.
# pylint:disable=invalid-name
outer_around = CallbackDecorator(CALLBACK_REGISTRY, 'around',
                                 priority_class=PriorityClass.DISPLAY)
# pylint:enable=invalid-name


# Global reference to the Terminal:
# This exists because the hooks have to be registered before the test is
# started, which is when the stream is passed in.
TERMINAL = [None]


class Terminal(blessings.Terminal):
    """
    Wrapped Terminal object for display hooks.

     * Adds additional features: write and writeln.
     * Adds a decorator to require the terminal global being set and
       pass it into hooks.
     * Works around a bug in blessings.Terminal.
    """

    def write(self, arg='', return_=False):
        """
        Write to the stream.

        :param return_: if True return the cursor back to where it was
            before the write.
        """
        self.stream.write(arg)

        if return_:
            self.stream.write(self.move_up * arg.count('\n'))

    def writeln(self, arg='', return_=False):
        """Convenience function to write a line to the stream."""
        self.write(arg + '\n', return_=return_)

    def color(self, color):  # pylint:disable=arguments-differ
        """Wrap Terminal.color() to work if styling is not set"""
        return super(Terminal, self).color(color) if self.does_styling else str

    @classmethod
    def required(cls, func):
        """Decorate this hook to only execute if term is available"""

        @wraps(func)
        def inner(*args, **kwargs):
            """Wrap func to check for term."""

            def empty_generator():
                """
                Hide the generator in a separate function
                because Python 2 can't support "returning from generators"
                """
                yield

            if not TERMINAL[0]:
                return empty_generator()
            else:
                return func(TERMINAL[0], *args, **kwargs)

        return inner


@outer_around.each_feature
@contextmanager
@Terminal.required
def feature_wrapper(term, feature):
    """Display feature execution."""

    try:
        if feature.tags:
            term.writeln(term.cyan(feature.represent_tags()))

        lines = feature.represented().splitlines()
        term.writeln(term.bold_white(lines[0]))
        term.writeln()
        term.writeln(term.white('\n'.join(lines[1:])))
        term.writeln()

        yield
    finally:
        pass


@outer_around.each_example
@contextmanager
@Terminal.required
def example_wrapper(term, scenario, outline, steps):
    """Display scenario execution."""

    try:
        if scenario.tags:
            term.writeln(term.cyan(scenario.represent_tags()))

        represented = scenario.represented()
        represented = ljust(represented, scenario.feature.max_length + 2)

        term.write(term.bold_white(represented))
        term.writeln(term.color(8)(scenario.location))

        if outline:
            term.writeln(represent_table([outline.keys(),
                                          outline.values()],
                                         indent=6,
                                         cell_wrap=term.white))
            term.writeln()

        # write a preview of the steps
        if term.is_a_tty:
            steps_ = []

            if scenario.feature.background:
                steps_ += scenario.feature.background.steps

            steps_ += steps

            term.writeln(
                term.color(8)('\n'.join(
                    step.represented()
                    for step in steps_
                ) + '\n'),
                return_=True
            )

        yield
    finally:
        term.writeln()


@outer_around.each_step
@contextmanager
@Terminal.required
@hook_not_reentrant  # don't display inner steps called by top-level steps
def step_wrapper(term, step):
    """Display step execution."""

    try:
        if term.is_a_tty:
            term.writeln(
                step.represented(color=term.color(11)),
                return_=True)

        yield
    finally:
        if step.passed:
            color = term.bold_green
        elif step.failed:
            color = term.bold_red
        else:
            color = term.yellow

        term.writeln(
            step.represented(color=color)
        )


class AloeTestResult(TextTestResult):
    """
    Cucumber test progress display (verbosity level 3).
    """
    # pylint:disable=too-many-arguments
    def __init__(self, stream, descriptions, verbosity,
                 config=None, errorClasses=None):
        super().__init__(stream, descriptions, verbosity,
                         config=config, errorClasses=errorClasses)
        self.showAll = verbosity == 2
        self.showSteps = verbosity >= 3  # pylint:disable=invalid-name

        if self.showSteps:
            # Register the global terminal so that it can be accessed by
            # the hooks.
            TERMINAL[0] = Terminal(
                stream=stream,
                force_styling=config.force_color)
        else:
            # unset the global -- primarily for the tests where we're going
            # to create more TestResults with different streams.
            TERMINAL[0] = None
